from .qzss_dcr_decoder_jma_common import QzssDcrDecoderJmaCommon
from ..definition import qzss_dcr_jma_typhoon_intensity_category
from ..definition import qzss_dcr_jma_typhoon_reference_time_type
from ..definition import qzss_dcr_jma_typhoon_scale_category
from ..exception import QzssDcrDecoderException
from ..report import QzssDcReportJmaBase
from ..report import QzssDcReportJmaTyphoon


class QzssDcrDecoderJmaTyphoon(QzssDcrDecoderJmaCommon):
    schema = QzssDcReportJmaBase

    def decode(self):
        self.reference_time = self.extract_day_hour_min_field(53)

        dt = self.extract_field(69, 3)
        try:
            self.reference_time_type = qzss_dcr_jma_typhoon_reference_time_type[dt]
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined JMA Type of Reference Time: {dt}',
                self)
        self.reference_time_type_raw = dt

        # the time elapsed from the analysis
        self.elapsed_time_from_reference_time = self.extract_field(80, 7)

        tn = self.extract_field(87, 7)
        if tn < 1 or tn > 99:
            raise QzssDcrDecoderException(
                f'Invalid JMA Typhoon Number: {tn}',
                self)
        self.typhoon_number = f'{tn}号'
        self.typhoon_number_raw = tn

        sr = self.extract_field(94, 4)
        try:
            self.typhoon_scale_category = qzss_dcr_jma_typhoon_scale_category[sr]
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined JMA Typhoon Scale Category: {sr}',
                self)
        self.typhoon_scale_category_raw = sr

        lc = self.extract_field(98, 4)
        try:
            self.typhoon_intensity_category = qzss_dcr_jma_typhoon_intensity_category[lc]
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined JMA Typhoon Intensity Category: {lc}',
                self)
        self.typhoon_intensity_category_raw = lc

        self.coordinates_of_typhoon = self.extract_lat_lon_field(102)

        pr = self.extract_field(143, 11)
        if pr > 1100:
            raise QzssDcrDecoderException(
                f'Invalid JMA Central Pressure: {pr}',
                self)
        self.central_pressure = f'{pr}hPa'
        self.central_pressure_raw = pr

        w1 = self.extract_field(154, 7)
        if w1 == 0:
            self.maximum_wind_speed = '不明'
        elif w1 < 15 or w1 > 105:
            raise QzssDcrDecoderException(
                f'Invalid JMA Maximum Wind Speed: {w1}',
                self)
        else:
            self.maximum_wind_speed = f'{w1}m/s'
        self.maximum_wind_speed_raw = w1

        w2 = self.extract_field(161, 7)
        if w2 == 0:
            self.maximum_gust_wind_speed = '不明'
        elif w2 < 15 or w2 > 105:
            raise QzssDcrDecoderException(
                f'Invalid JMA Maximum Gust Wind Speed: {w2}',
                self)
        else:
            self.maximum_gust_wind_speed = f'{w2}m/s'
        self.maximum_gust_wind_speed_raw = w2

        return QzssDcReportJmaTyphoon(**self.get_params())



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcr_jma_long_period_ground_motion_upper_limit = QzssDcrDefinition(
    {
        0: None,
        1: "長周期地震動階級1未満",
        2: "長周期地震動階級1",
        3: "長周期地震動階級2",
        4: "長周期地震動階級3",
        5: "長周期地震動階級4",
        6: "〜程度以上",
        7: "不明",
    },
    undefined="長周期地震動階級(上限)(コード番号：%d)"
)



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcr_jma_volcanic_warning_code = QzssDcrDefinition(
    {
        11: "レベル1(活火山であることに留意)",
        12: "レベル2(火口周辺規制)",
        13: "レベル3(入山規制)",
        14: "レベル4(高齢者等避難)",
        15: "レベル5(避難)",
        21: "活火山であることに留意",
        22: "火口周辺危険",
        23: "入山危険",
        24: "山麓厳重警戒",
        25: "居住地域厳重警戒",
        35: "活火山であることに留意(海底火山)",
        36: "周辺海域警戒",
        52: "噴火",
        62: "噴火したもよう",
        127: "その他の防災気象情報要素",
        # "NNN*": "防災気象情報要素(コード番号：NNN)",
    },
    undefined="防災気象情報要素(コード番号：%d)"
)



from .qzss_dcr_decoder_jma_common import QzssDcrDecoderJmaCommon
from ..definition import qzss_dcr_jma_information_serial_code
from ..exception import QzssDcrDecoderException
from ..report import QzssDcReportJmaBase
from ..report import QzssDcReportJmaNankaiTroughEarthquake


class QzssDcrDecoderJmaNankaiTroughEarthquake(QzssDcrDecoderJmaCommon):
    schema = QzssDcReportJmaBase

    def decode(self):
        ie = self.extract_field(53, 4)
        try:
            self.information_serial_code = qzss_dcr_jma_information_serial_code[ie]
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined JMA Information Serial Code: {ie}',
                self)
        self.information_serial_code_raw = ie

        te = []
        for i in range(18):
            te.append(self.extract_field(57 + i * 8, 8))

        self.text_information = bytes(te)
        self.page_number = self.extract_field(201, 6)
        self.total_page = self.extract_field(207, 6)

        return QzssDcReportJmaNankaiTroughEarthquake(**self.get_params())



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcr_jma_flood_warning_level = QzssDcrDefinition(
    {
        1: "警報解除",
        2: "氾濫警戒情報",
        3: "氾濫危険情報",
        4: "氾濫発生情報",
        15: "その他の警戒レベル",
        # "NN*": "警戒レベル(コード番号：NN)",
    },
    undefined="警戒レベル(コード番号：%d)"
)



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcr_jma_report_classification = QzssDcrDefinition(
    {
        1: "最優先",
        2: "優先",
        3: "通常",
        7: "訓練/試験",
    },
    undefined="通報区分(コード番号：%d)"
)

qzss_dcr_jma_report_classification_en = QzssDcrDefinition(
    {
        1: "Maximum Priority",
        2: "Priority",
        3: "Regular",
        7: "Training/Test",
    },
    undefined="Undefined Report Classification (Code: %d)"
)



from .qzss_dcr_definition import QzssDcrDefinition


qzss_dcx_camf_a3_provider_identifier_australia = QzssDcrDefinition(
    {
        1: 'National Emergency Management Agency',
        2: 'Bureau of Meteorology',
        3: 'Australian Climate Service',
        4: 'Geoscience Australia',
        5: 'Commonwealth Scientific and Industrial Research Organisation',
        6: 'Australian Bureau of Statistics',
        7: 'Resilience New South Wales',
        8: 'State Emergency Service New South Wales',
        9: 'New South Wales Rural Fire Service',
        10: 'Joint Australian Tsunami Warning Centre',
        11: 'Flood Knowledge Centre',
        12: 'Australian Broadcasting Corporation',
    },
    undefined="Undefined Provider Identifier of Australia (Code: %d)"
)


qzss_dcx_camf_a3_provider_identifier_fiji = QzssDcrDefinition(
    {
        1: 'National Disaster Management Office',
        2: 'Fiji Meteorological Service',
        3: 'Hydrology Section, Fiji Water Authority',
        4: 'Mineral Resources Department',
        5: 'Fiji Broadcasting Corporation',
    },
    undefined="Undefined Provider Identifier of Fiji (Code: %d)"
)


qzss_dcx_camf_a3_provider_identifier_japan = QzssDcrDefinition(
    {
        1: 'Foundation for MultiMedia Communications',
        2: 'Fire and Disaster Management Agency',
        3: 'Related Ministries',
        4: 'Local Government',
    },
    undefined="Undefined Provider Identifier of Japan (Code: %d)"
)


qzss_dcx_camf_a3_provider_identifier_thailand = QzssDcrDefinition(
    {
        1: 'Department of Disaster Prevention and Mitigation',
        2: 'Thai Meteorological Department',
        3: 'National Disaster Warning Center',
        4: 'Department of Mineral Resources',
        5: 'Navy Hydrographic Department, Royal Thai Navy',
        6: 'Department of Water Resources',
        7: 'Royal Irrigation Department',
        8: 'Department of Pollution Control',
        9: 'Geo-Informatics and Space Technology Development Agency',
        10: 'Electricity Generating Authority of Thailand',
        11: 'Royal Forest Department',
        12: 'Department of Parks, Wildlife and Plant Conservation',
        13: 'Water Crisis Prevention Center',
    },
    undefined="Undefined Provider Identifier of Thailand (Code: %d)"
)


qzss_dcx_camf_a3_provider_identifier_undefined = QzssDcrDefinition(
    {},
    undefined="Undefined Provider Identifier (Code: %d)"
)


qzss_dcx_camf_a3_provider_identifier_map = {
    10: qzss_dcx_camf_a3_provider_identifier_australia,
    71: qzss_dcx_camf_a3_provider_identifier_fiji,
    111: qzss_dcx_camf_a3_provider_identifier_japan,
    219: qzss_dcx_camf_a3_provider_identifier_thailand,
}



from .qzss_dcr_decoder_jma_common import QzssDcrDecoderJmaCommon
from ..definition import qzss_dcr_jma_tsunami_forecast_region
from ..definition import qzss_dcr_jma_tsunami_height
from ..definition import qzss_dcr_jma_tsunami_warning_code
from ..exception import QzssDcrDecoderException
from ..report import QzssDcReportJmaBase
from ..report import QzssDcReportJmaTsunami


class QzssDcrDecoderJmaTsunami(QzssDcrDecoderJmaCommon):
    schema = QzssDcReportJmaBase

    def decode(self):
        self.notifications_on_disaster_prevention, self.notifications_on_disaster_prevention_raw =\
            self.extract_notification_on_disaster_prevention_fields(53)
        dw = self.extract_field(80, 4)
        try:
            self.tsunami_warning_code = qzss_dcr_jma_tsunami_warning_code[dw]
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined JMA Tsunami Warning Code: {dw}',
                self)
        self.tsunami_warning_code_raw = dw

        self.expected_tsunami_arrival_times = []
        self.tsunami_heights = []
        self.tsunami_heights_raw = []
        self.tsunami_forecast_regions = []
        self.tsunami_forecast_regions_raw = []
        for i in range(5):
            offset = 84 + i * 26
            if self.extract_field(offset, 26) == 0:
                break

            self.expected_tsunami_arrival_times.append(self.extract_expected_tsunami_arrival_time(offset))

            th = self.extract_field(offset + 12, 4)
            try:
                self.tsunami_heights.append(qzss_dcr_jma_tsunami_height[th])
            except KeyError:
                raise QzssDcrDecoderException(
                    f'Undefined JMA Tsunami Height: {th}',
                    self)
            self.tsunami_heights_raw.append(th)

            pl = self.extract_field(offset + 16, 10)
            try:
                self.tsunami_forecast_regions.append(qzss_dcr_jma_tsunami_forecast_region[pl])
            except KeyError:
                raise QzssDcrDecoderException(
                    f'Undefined JMA Tsunami Forecast Region: {pl}',
                    self)
            self.tsunami_forecast_regions_raw.append(pl)

        return QzssDcReportJmaTsunami(**self.get_params())



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcr_jma_seismic_intensity_lower_limit = QzssDcrDefinition(
    {
        1: "震度0",
        2: "震度1",
        3: "震度2",
        4: "震度3",
        5: "震度4",
        6: "震度5弱",
        7: "震度5強",
        8: "震度6弱",
        9: "震度6強",
        10: "震度7",
        14: "なし",
        15: "不明",
    },
    undefined="震度(下限)(コード番号：%d)"
)



import enum
from .qzss_dcr_definition import QzssDcrDefinition


class DcxMessageType(enum.Enum):
    NULL_MSG = 0
    OUTSIDE_JAPAN = 1
    L_ALERT = 2
    J_ALERT = 3
    LOCAL_GOV = 4
    UNKNOWN = 5


qzss_dcx_message_type = QzssDcrDefinition(
    {
        DcxMessageType.NULL_MSG: 'Null Message',
        DcxMessageType.OUTSIDE_JAPAN: 'Information from Organizations outside Japan',
        DcxMessageType.L_ALERT: 'L-Alert',
        DcxMessageType.J_ALERT: 'J-Alert',
        DcxMessageType.LOCAL_GOV: 'Information from Local Government',
        DcxMessageType.UNKNOWN: 'Unknown DCX Message Type',
    },
    undefined='Undefined Message Type (Code: %d)'
)


import os
from pprint import pprint

import pytest

import azarashi


def test_scenario1():
    # 緊急地震速報
    report = azarashi.decode('$QZQSM,58,9AAF899C80000324000039000548C5E2C000000003DFF8001C000012FE4B0FC*7F')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaEarthquakeEarlyWarning
    # 震度
    report = azarashi.decode('$QZQSM,58,C6AF999C828001C82CB25AE775A8D4CA854AB8000000000000000011E027E5C*76')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaSeismicIntensity
    # 震源
    report = azarashi.decode('$QZQSM,58,9AAF919C82800388000039051440C5C82A0108300000000000000012497DA18*0A')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaHypocenter
    # 津波
    report = azarashi.decode('$QZQSM,58,9AAFA99C828001E8F67C31053960414E621053BE00000000000000132735038*0F')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaTsunami


def test_scenario2():
    # 津波(更新)
    report = azarashi.decode('$QZQSM,58,9AAFA99C8C8001E8F67C31193960464E621193BBC464EF80000000109DB7028*09')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaTsunami


def test_scenario3():
    # 南海トラフ地震
    report = azarashi.decode('$QZQSM,58,C6AFA19C918002F2C6CBF35ADBF1C1C471C1D4F1C1CAF3595F82D81262EF438*02')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,9AAFA19C918002F1C0D471C0D1F1C0D371C0C7F1C0D071C0CAB4D812D45BCD0*06')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,9AAFA19C918002F1C0CFF4CC5973C0DEF2D7DF72DFCE71C14932D813A64867C*0C')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,53AFA19C918002F1C0C5F1C0C8F3D6C4F1C0D5F2DFCE71C0CC30D81029188EC*0F')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,C6AFA19C918002F1C0D4F1C0C5F1C144F1C0D772C8DE71C0D9AED8108F361A8*7A')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,9AAFA19C918002F2DD4E71C1427443D573595DF25EC9F1C0D52CD81113EFC54*00')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,9AAFA19C918002F1C0CCF1C041725DC572DF4671C0D7734A5FAAD811D504B9C*09')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,53AFA19C918002F1C0D4744041F1C0C471C144F1C14671C0DF28D8127EDCCCC*02')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,C6AFA19C918002F4D5CC71C0DF71C0D1F1C0D371C0C271C145A6D812E9F0C70*03')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,9AAFA19C918002F1C0DCF1C0D373CDDC72D7DF73CD4271C0D5A4D813579A170*78')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,9AAFA19C918002F1C0C672DCD9F2DC5C734CC171C0D5F357CA22D8100D3A898*07')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,53AFA19C918002F1C0D773CCDD73CA4FF2C7D7F441DEF34053A0D8108524298*01')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,C6AFA19C918002F1C040F2D253F45347F35450F2CE5874CE439ED8113990640*7D')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,9AAFA19C918002F2D74D74CE43F35D4872CFCFF1C0D3F1C0D79CD8118D63AC8*07')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,9AAFA19C918002F1C1D4F1C1CAF2CE5874CE43F1C0D77341D99AD81268C2E7C*73')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,53AFA19C918002F1C14471C14571C040F2C6CBF35ADBF1C1C418D812DC94FA0*7F')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,C6AFA19C918002F1C0DCF1C14571C0D773CCDD73CA4FF1C0D596D81357F75D0*74')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,9AAFA19C918002F1C145F1C14371C0D1F1C0C7F1C14571C0CC94D813ED6741C*0C')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,9AAFA19C918002F4C04D72DC5C71C0D471C0D7F3CAD871C0D512D810ADB00F0*78')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,53AFA19C918002F1C14671C0DF71C0CCF1C04171C0C9F1C0D710D8112F31B60*73')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,C6AFA19C918002F1C0C9F1C0D471C0C673475472D74D71C0CA8ED8118DCA59C*7E')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,9AAFA19C918002F3CCDD73CA4FF1C0CBF1C0D371C0C271C1458CD81215E2B50*08')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,9AAFA19C918002F1C0C7F1C14571C0CCF1C0DCF1C14571C0C60AD812CB85F9C*06')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,53AFA19C918002F1C0D7F3CAD871C0D571C145F1C14371C0D188D81363A8A1C*7E')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,C6AFA19C918002F2D141F3CAC671C0D3F4C04D72DC5C71C0D406D813ECF2998*05')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    report = azarashi.decode('$QZQSM,58,9AAFA19C918002F1C0C271C0D771C1CBF1C1D671C1DE71C1C404D8104D300D8*7F')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    assert report.completed is False
    report = azarashi.decode('$QZQSM,58,9AAFA19C918002F1C0C271C04100000000000000000000000036D81121AA2D0*07')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNankaiTroughEarthquake
    assert report.completed is True


def test_scenario4():
    # 火山
    report = azarashi.decode('$QZQSM,58,C6AFC19CA50001CA5341F783E0F10910421230200000000000000011B086438*70')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaVolcano
    # 降灰(速報)
    report = azarashi.decode('$QZQSM,58,9AAFC99CA50001CA523EE4C1F07826122081309181000000000000121BAF1C0*71')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaAshFall


def test_scenario5():
    # 降灰(詳細)
    report = azarashi.decode('$QZQSM,58,C6AFC99CAA0001CAA43EE541F0782A1220813091811183E0F000001329B16E0*7F')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaAshFall
    report = azarashi.decode('$QZQSM,58,C6AFC99CAA0001CAA43EE8C2441046123023307C1E19848820000013E2F3E6C*00')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaAshFall
    report = azarashi.decode('$QZQSM,58,9AAFC99CAA0001CAA43EECC24604860F83C430910421848C080000106286874*00')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaAshFall
    report = azarashi.decode('$QZQSM,58,53AFC99CAA0001CAA43EF4C1F078A61220853091813183E0F0000010CE8C39C*77')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaAshFall
    report = azarashi.decode('$QZQSM,58,C6AFC99CAA0001CAA43EF8C24410C61230200000000000000000001148565F4*0A')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaAshFall


def test_scenario6():
    # 気象
    report = azarashi.decode('$QZQSM,58,C6AFD19CB18001113880115F901186A011ADB011D4C011FBD00000135EAA3F8*73')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaWeather
    report = azarashi.decode('$QZQSM,58,9AAFD19CB180011222E0B93880B95F90B986A0B9ADB0B9D4C0000013D276B60*0D')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaWeather
    report = azarashi.decode('$QZQSM,58,53AFD19CB18001B9FBD0BA22E00000000000000000000000000000107AA71EC*76')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaWeather
    # 洪水
    report = azarashi.decode('$QZQSM,58,C6AFD99CB1800160A8F5528600000000000000000000000000000010E502538*0E')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaFlood


def test_scenario7():
    # 洪水(警報解除)
    report = azarashi.decode('$QZQSM,58,C6AFD99CB90000E0A8F5528600000000000000000000000000000013A699D5C*76')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaFlood


def test_hex():
    # Hexadecimal
    azarashi.decode('C6AF89A820000324000050400548C5E2C000000003DFF8001C00001185443FC', 'hex')


def test_ublox():
    # Ublox SFRBX
    report_ublox = azarashi.decode(
            b'\xB5\x62\x02\x13\x2C\x00\x05\x02\x01\x00\x09\x40\x02\x00\xC5\xF1'
            b'\xAD\x9A\x04\x05\x80\x11\x54\x8D\xA0\x60\x3F\x82\xD2\x11\x0F\xAA'
            b'\x7D\x50\x28\x0C\x43\xC9\x10\x00\x50\x7D\x31\x79\xF0\x28\x73\x18'
            b'\x10\xB2\x62\x2F',
            'ublox')
    report_nmea = azarashi.decode(report_ublox.nmea, 'nmea')
    assert report_ublox.nmea == report_nmea.nmea
    assert report_ublox.satellite_id == report_nmea.satellite_id == 56
    assert report_ublox.satellite_prn == report_nmea.satellite_prn == 184

    report_ublox = azarashi.decode(
            b'\xB5\x62\x02\x13\x2C\x00\x05\x02\x01\x00\x09\x40\x02\x00\x76\x14'
            b'\xAD\x53\x5C\x03\x80\x1A\x33\xEC\x00\x00\x49\x48\x2F\x14\x20\x1B'
            b'\x01\x52\x00\x00\x00\x00\x12\x00\x00\x00\xA6\x59\xB0\xC6\x5B\x1F'
            b'\xF3\xB2\x3E\xB6',
            'ublox')
    report_nmea = azarashi.decode(report_ublox.nmea, 'nmea')
    assert report_ublox.nmea == report_nmea.nmea
    assert report_ublox.satellite_id == report_nmea.satellite_id == 56
    assert report_ublox.satellite_prn == report_nmea.satellite_prn == 184

    report_ublox = azarashi.decode(
            b'\xb5\x62\x02\x13\x2c\x00\x05\x03\x01\x00\x09\x41\x02\x00\xa1\xe5'
            b'\xad\xc6\x12\x02\x80\x36\x68\x00\x00\x01\x10\x50\x34\x44\xd4\xee'
            b'\x00\x34\x00\x00\x00\x3c\x11\x00\x00\x00\x14\x9c\x5f\x60\x26\xcb'
            b'\xc0\xf2\xef\x1a',
            'ublox')
    
    report_nmea = azarashi.decode(report_ublox.nmea, 'nmea')
    assert report_ublox.nmea == report_nmea.nmea
    assert report_ublox.satellite_id == report_nmea.satellite_id == 57
    assert report_ublox.satellite_prn == report_nmea.satellite_prn == 185
    
    report_ublox = azarashi.decode(
            b'\xb5\x62\x02\x13\x2c\x00\x05\x07\x01\x00\x09\x45\x02\x00\xa1\xe5'
            b'\xad\xc6\x12\x02\x80\x36\x68\x00\x00\x01\x10\x50\x34\x44\xd4\xee'
            b'\x00\x34\x00\x00\x00\x3c\x11\x00\x00\x00\x14\x9c\x5f\x60\x26\xcb'
            b'\xc0\xf2\xf7\x62',
            'ublox')
    
    report_nmea = azarashi.decode(report_ublox.nmea, 'nmea')
    assert report_ublox.nmea == report_nmea.nmea
    assert report_ublox.satellite_id == report_nmea.satellite_id == 61
    assert report_ublox.satellite_prn == report_nmea.satellite_prn == 189
    
    report_ublox = azarashi.decode( # svid:1 は停波済
            b'\xb5\x62\x02\x13\x2c\x00\x05\x01\x01\x00\x09\x45\x02\x00\xa1\xe5'
            b'\xad\xc6\x12\x02\x80\x36\x68\x00\x00\x01\x10\x50\x34\x44\xd4\xee'
            b'\x00\x34\x00\x00\x00\x3c\x11\x00\x00\x00\x14\x9c\x5f\x60\x26\xcb'
            b'\xc0\xf2\xf1\x60',
            'ublox')
    
    report_nmea = azarashi.decode(report_ublox.nmea, 'nmea')
    assert report_ublox.nmea == report_nmea.nmea
    assert report_ublox.satellite_id == report_nmea.satellite_id == 55
    assert report_ublox.satellite_prn == report_nmea.satellite_prn == 183


def test_northwest_pacific_tsunami():
    # 北西太平洋津波
    report = azarashi.decode('$QZQSM,55,53AD360D5B80047FFFFE3000000000000000000000000000000000118372EC8*0C')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNorthwestPacificTsunami
    report = azarashi.decode('$QZQSM,56,9AAD3609E080023AE008D3D1008E449009D457009E3E5011F00000138B3E720*09')
    assert type(report) == azarashi.qzss_dc_report.QzssDcReportJmaNorthwestPacificTsunami


def test_unknown_magnitude_or_depth():
    # 震源 マグニチュード:不明, 深さ:不明
    report = azarashi.decode('$QZQSM,55,53AD160D2800039400001A28FFFFEE601800C8F00000000000000011BF8D908*01')
    assert report.depth_of_hypocenter == '不明'
    assert report.depth_of_hypocenter_raw == 511
    assert report.magnitude == '不明'
    assert report.magnitude_raw == 127


def test_long_period_ground_motion():
    # 緊急地震速報 長周期地震動
    report = azarashi.decode('$QZQSM,56,9AAF88A48000DB24000049000548C5E2C000000003DFF8001C000012101445C*7B')
    assert report.long_period_ground_motion_lower_limit == '長周期地震動階級2'
    assert report.long_period_ground_motion_lower_limit_raw == 3
    assert report.long_period_ground_motion_upper_limit == '長周期地震動階級2'
    assert report.long_period_ground_motion_upper_limit_raw == 3


def test_from_file():
    dir_path = os.path.dirname(os.path.realpath(__file__))

    def handler(report):
        pprint(report.get_params())

    with open(dir_path + '/qzqsm_220307.log', mode='r', encoding='utf-8') as f:
        while True:
            try:
                azarashi.decode_stream(f, msg_type='nmea', callback=handler)
            except EOFError as e:
                print(f'# [{type(e).__name__}] {e}')
                break


def test_decode_error():
    with pytest.raises(azarashi.QzssDcrDecoderException):
        # crc mismatch
        azarashi.decode('C6AF89A820000324000050400548C5E2C000000003DFF8001C000011854432D', 'hex')

    with pytest.raises(azarashi.QzssDcrDecoderException):
        # checksum mismatch
        azarashi.decode('$QZQSM,55,C6AF89A820000324000050400548C5E2C000000003DFF8001C00001185443FC*00')

    with pytest.raises(azarashi.QzssDcrDecoderException):
        # checksum mismatch
        azarashi.decode(b'\xB5\x62\x02\x13\x2C\x00\x05\x02\x01\x00\x09\x40\x02\x00\xC5\xF1'
                        b'\xAD\x9A\x04\x05\x80\x11\x54\x8D\xA0\x60\x3F\x82\xD2\x11\x0F\xAA'
                        b'\x7D\x50\x28\x0C\x43\xC9\x10\x00\x50\x7D\x31\x79\xF0\x28\x73\x18'
                        b'\x10\xB2\x62\x20',
                        'ublox')


def test_dcx():
    null_msg = azarashi.decode('$QZQSM,55,53B0840DE0000000000000000000000000000000000000000000000012ACBD4*0E')
    assert type(null_msg) is azarashi.qzss_dc_report.QzssDcxNullMsg

    org_outside_jpn = azarashi.decode('$QZQSM,56,9AB08408E0598969E00066AFFE8E6F70091200000000000000000100CD1A410*0C')
    assert type(org_outside_jpn) is azarashi.qzss_dc_report.QzssDcxOutsideJapan

    l_alert = azarashi.decode('$QZQSM,55,9AB0840DE10208ADE0000000000000000000011340000000000000132F0D238*04')
    assert type(l_alert) is azarashi.qzss_dc_report.QzssDcxLAlert

    j_alert = azarashi.decode('$QZQSM,55,53B0840DE31188FC208600000000000000001FFFFFFFFFFFC00000120738628*00')
    assert type(j_alert) is azarashi.qzss_dc_report.QzssDcxJAlert



def hex_qzss_dcr_message_extractor(reader, reader_args=(), reader_kwargs={}):
    msg = reader(*reader_args, **reader_kwargs)
    if not msg:
        raise EOFError('Encountered EOF')
    elif type(msg) == bytes or type(msg) == bytearray:
        msg = msg.decode()
    return msg



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcr_jma_seismic_intensity = QzssDcrDefinition(
    {
        1: "4未満",
        2: "4",
        3: "5弱",
        4: "5強",
        5: "6弱",
        6: "6強",
        7: "7",
    },
    undefined="震度(コード番号：%d)"
)



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcr_jma_typhoon_intensity_category = QzssDcrDefinition(
    {
        0: "なし",
        1: "強い",
        2: "非常に強い",
        3: "猛烈な",
        15: "その他の強さ階級分類",
        # "NN*": "強さ階級分類(コード番号：NN)",
    },
    undefined="強さ階級分類(コード番号：%d)"
)



import datetime
import math

from .qzss_dcr_decoder_base import QzssDcrDecoderBase
from ..definition import qzss_dcx_camf_a10_library_version
from ..definition import qzss_dcx_camf_a11_international_library
from ..definition import qzss_dcx_camf_a11_international_library_code
from ..definition import qzss_dcx_camf_a11_japanese_library_en
from ..definition import qzss_dcx_camf_a11_japanese_library_ja
from ..definition import qzss_dcx_camf_a17_main_subject_for_specific_settings
from ..definition import qzss_dcx_camf_a1_message_type
from ..definition import qzss_dcx_camf_a2_country_region_name
from ..definition import qzss_dcx_camf_a3_provider_identifier_map
from ..definition import qzss_dcx_camf_a3_provider_identifier_undefined
from ..definition import qzss_dcx_camf_a4_hazard_category
from ..definition import qzss_dcx_camf_a4_hazard_definition
from ..definition import qzss_dcx_camf_a4_hazard_type
from ..definition import qzss_dcx_camf_a5_severity
from ..definition import qzss_dcx_camf_a6_hazard_onset_week
from ..definition import qzss_dcx_camf_a8_hazard_duration
from ..definition import qzss_dcx_camf_a9_selection_of_library
from ..definition import qzss_dcx_camf_c10_guidance_library_for_second_ellipse
from ..definition import qzss_dcx_camf_c10_guidance_library_for_second_ellipse_code
from ..definition.qzss_dcx_camf_b4_lower_level_fields_tables import *
from ..definition.qzss_dcx_camf_ex1_target_area_code import *
from ..definition.qzss_dcx_camf_ex9_target_area_code import *
from ..definition.qzss_dcx_message_type import *
from ..report import QzssDcReportMessageBase, QzssDcxNullMsg, QzssDcxOutsideJapan, QzssDcxLAlert, QzssDcxJAlert, \
    QzssDcxMTInfo, QzssDcxUnknown


class QzssDcxDecoder(QzssDcrDecoderBase):
    schema = QzssDcReportMessageBase

    def decode(self):
        class CAMF:
            def __str__(self):
                return str(self.get_params())

            def get_params(self):
                return self.__dict__

        self.camf = camf = CAMF()
        camf.sdmt = self.extract_field(14, 1)
        camf.sdm = self.extract_field(15, 9)
        camf.a1 = self.extract_field(24, 2)
        camf.a2 = self.extract_field(26, 9)
        camf.a3 = self.extract_field(35, 5)
        camf.a4 = self.extract_field(40, 7)
        camf.a5 = self.extract_field(47, 2)
        camf.a6 = self.extract_field(49, 1)
        camf.a7 = self.extract_field(50, 14)
        camf.a8 = self.extract_field(64, 2)
        camf.a9 = self.extract_field(66, 1)
        camf.a10 = self.extract_field(67, 3)
        camf.a11 = self.extract_field(70, 10)
        camf.a12 = self.extract_field(80, 16)
        camf.a13 = self.extract_field(96, 17)
        camf.a14 = self.extract_field(113, 5)
        camf.a15 = self.extract_field(118, 5)
        camf.a16 = self.extract_field(123, 6)
        camf.a17 = self.extract_field(129, 2)
        camf.a18 = self.extract_field(131, 15)
        camf.ex1 = self.extract_field(146, 16)
        camf.ex2 = self.extract_field(162, 1)
        camf.ex3 = self.extract_field(163, 17)
        camf.ex4 = self.extract_field(180, 17)
        camf.ex5 = self.extract_field(197, 5)
        camf.ex6 = self.extract_field(202, 5)
        camf.ex7 = self.extract_field(207, 7)
        camf.ex8 = self.extract_field(146, 1)
        camf.ex9 = self.extract_field(147, 64)
        camf.ex10 = self.extract_field(211, 3)
        camf.vn = self.extract_field(214, 6)

        if camf.sdmt == 0:
            self.satellite_designation_mask_type = 'MT44 is for Japan or for use outside Japan'
            sd_list = ['For Japan', 'For use outside Japan']
        else:  # sdmt should be 1.
            self.satellite_designation_mask_type = 'MT44 transmission status'
            sd_list = ['Transmission stopped', 'Transmission in progress']

        self.satellite_designation_mask = []
        for i in range(9):
            self.satellite_designation_mask.append(sd_list[(camf.sdm & 1 << i) >> i])

        if camf.a1 == camf.a3 == camf.a4 == camf.a5 == camf.a6 == camf.a7 == camf.a8 == camf.a9 == camf.a10 == \
                camf.a11 == camf.a12 == camf.a13 == camf.a14 == camf.a15 == camf.a16 == camf.a17 == camf.a18 == \
                camf.ex1 == camf.ex2 == camf.ex3 == camf.ex4 == camf.ex5 == camf.ex6 == camf.ex7 == camf.vn == 0:
            dcx_message_type = DcxMessageType.NULL_MSG
        elif camf.a2 == 111:  # japan
            if camf.a3 == 1:  # fmmc
                dcx_message_type = DcxMessageType.L_ALERT
            elif camf.a3 == 2 or camf.a3 == 3:  # fdma or related ministries
                dcx_message_type = DcxMessageType.J_ALERT
            elif camf.a3 == 4:  # information from local government
                dcx_message_type = DcxMessageType.LOCAL_GOV
            else:
                dcx_message_type = DcxMessageType.UNKNOWN
        else:  # outside japan
            dcx_message_type = DcxMessageType.OUTSIDE_JAPAN

        # detecting the message type
        self.dcx_message_type = qzss_dcx_message_type[dcx_message_type]
        # setting flags to ignore fields
        self.ignore_a12_to_a16 = False
        self.ignore_a17_to_a18 = False
        self.ignore_ex1 = False
        self.ignore_ex2_to_ex7 = False
        self.ignore_ex8_to_ex9 = False
        if dcx_message_type == DcxMessageType.OUTSIDE_JAPAN:
            self.ignore_ex1 = True
            self.ignore_ex2_to_ex7 = True
            self.ignore_ex8_to_ex9 = True
        elif dcx_message_type == DcxMessageType.L_ALERT:
            a12_to_a16 = self.extract_field(80, 49)
            if a12_to_a16 == 0:
                self.ignore_a12_to_a16 = True
            else:
                self.ignore_ex1 = True
            self.ignore_ex2_to_ex7 = True
            self.ignore_ex8_to_ex9 = True
        elif dcx_message_type == DcxMessageType.J_ALERT:
            self.ignore_a12_to_a16 = True
            self.ignore_a17_to_a18 = True
            self.ignore_ex1 = True
            self.ignore_ex2_to_ex7 = True
        elif dcx_message_type == DcxMessageType.LOCAL_GOV:
            ex3_to_ex7 = self.extract_field(163, 51)
            if ex3_to_ex7 == 0:
                self.ignore_ex2_to_ex7 = True
            self.ignore_ex8_to_ex9 = True
        elif dcx_message_type == DcxMessageType.NULL_MSG:
            self.ignore_a12_to_a16 = True
            self.ignore_a17_to_a18 = True
            self.ignore_ex1 = True
            self.ignore_ex2_to_ex7 = True
            self.ignore_ex8_to_ex9 = True
            return QzssDcxNullMsg(**self.get_params())
        else:  # has to be an unknown message
            self.ignore_ex1 = True
            self.ignore_ex2_to_ex7 = True
            self.ignore_ex8_to_ex9 = True

        self.a1_message_type = qzss_dcx_camf_a1_message_type[camf.a1]
        self.a2_country_region_name = qzss_dcx_camf_a2_country_region_name[camf.a2]
        self.a3_provider_identifier = (
            qzss_dcx_camf_a3_provider_identifier_map.get(
                camf.a2,
                qzss_dcx_camf_a3_provider_identifier_undefined))[camf.a3]
        self.a4_hazard_category = qzss_dcx_camf_a4_hazard_category[camf.a4]
        self.a4_hazard_type = qzss_dcx_camf_a4_hazard_type[camf.a4]
        self.a4_hazard_definition = qzss_dcx_camf_a4_hazard_definition[camf.a4]
        self.a5_severity = qzss_dcx_camf_a5_severity[camf.a5]
        self.a6_hazard_onset_week = qzss_dcx_camf_a6_hazard_onset_week[camf.a6]
        today = datetime.datetime.now(datetime.UTC).replace(hour=0, minute=0, second=0, microsecond=0)
        hazard_onset = today + datetime.timedelta(weeks=camf.a6, days=-today.weekday(), minutes=camf.a7 - 1)
        if 1 <= camf.a7 <= 10080:
            self.a7_hazard_onset_time_of_week = hazard_onset.strftime('%A - %I:%M %p').upper()
            self.a6a7_hazard_onset_datetime = hazard_onset
        else:  # a7 is unexpectedly out of range...
            self.a7_hazard_onset_time_of_week = 'The Time of the Week Value is Unexpectedly Out of Range (Code: %d)' % camf.a7
            self.a6a7_hazard_onset_datetime = None
        self.a8_hazard_duration = qzss_dcx_camf_a8_hazard_duration[camf.a8]
        self.a9_selection_of_library = qzss_dcx_camf_a9_selection_of_library[camf.a9]
        self.a10_library_version = qzss_dcx_camf_a10_library_version[camf.a10]
        if camf.a10 == 0:  # library version #1 is only supported for now.
            if camf.a9 == 0:  # international library
                self.a11_international_library_code = qzss_dcx_camf_a11_international_library_code[camf.a11]
                self.a11_international_library = qzss_dcx_camf_a11_international_library[camf.a11]
            elif camf.a9 == 1:  # country/region guidance library
                if camf.a2 == 111:  # japanese library
                    self.a11_japanese_library = qzss_dcx_camf_a11_japanese_library_en[camf.a11]
                    self.a11_japanese_library_ja = qzss_dcx_camf_a11_japanese_library_ja[camf.a11]

        def __get_axis(factor):
            return pow(10, math.log10(216.2) + factor * (math.log10(2500000) - math.log10(216.2)) / 31) / 1000

        if self.ignore_a12_to_a16 is False:
            self.a12_ellipse_centre_latitude = -90 + 180 / 0xFFFF * camf.a12
            self.a13_ellipse_centre_longitude = -180 + 360 / 0x1FFFF * camf.a13
            self.a14_ellipse_semi_major_axis = __get_axis(camf.a14)
            self.a15_ellipse_semi_minor_axis = __get_axis(camf.a15)
            self.a16_ellipse_azimuth = -90 + 180 / 0x40 * camf.a16
        if self.ignore_a17_to_a18 is False:
            if camf.a17 == 0 and self.ignore_a12_to_a16 is False:  # B1 – improved resolution of main ellipse
                self.a17_main_subject_for_specific_settings = \
                    qzss_dcx_camf_a17_main_subject_for_specific_settings[camf.a17]
                camf.c1 = self.extract_field(131, 3)
                camf.c2 = self.extract_field(134, 3)
                camf.c3 = self.extract_field(137, 3)
                camf.c4 = self.extract_field(140, 3)
                self.c1_refined_latitude_of_centre_of_main_ellipse = (
                        self.a12_ellipse_centre_latitude +
                        (180 / 0xFFFF) / 8 * camf.c1
                )
                self.c2_refined_longitude_of_centre_of_main_ellipse = (
                        self.a13_ellipse_centre_longitude +
                        (180 / 0xFFFF) / 8 * camf.c2
                )
                if camf.a14 == 0:
                    delta = self.a14_ellipse_semi_major_axis
                else:
                    delta = self.a14_ellipse_semi_major_axis - __get_axis(camf.a14 - 1)
                self.c3_refined_length_of_semi_major_axis = (
                        self.a14_ellipse_semi_major_axis -
                        delta * camf.c3 / 8
                )
                if camf.a15 == 0:
                    delta = self.a15_ellipse_semi_minor_axis
                else:
                    delta = self.a15_ellipse_semi_minor_axis - __get_axis(camf.a15 - 1)
                self.c4_refined_length_of_semi_minor_axis = (
                        self.a15_ellipse_semi_minor_axis -
                        delta * camf.c4 / 8
                )
            elif camf.a17 == 1 and self.ignore_a12_to_a16 is False:  # B2 – position of centre of hazard
                self.a17_main_subject_for_specific_settings = \
                    qzss_dcx_camf_a17_main_subject_for_specific_settings[camf.a17]
                camf.c5 = self.extract_field(131, 7)
                camf.c6 = self.extract_field(138, 7)
                if camf.c5 <= 63:
                    delta = -10 + 20 / 0x80 * camf.c5
                else:
                    delta = -10 + 20 / 0x80 * (camf.c5 + 1)
                self.c5_latitude_of_centre_of_hazard = self.a12_ellipse_centre_latitude + delta
                if camf.c6 <= 63:
                    delta = -10 + 20 / 0x80 * camf.c6
                else:
                    delta = -10 + 20 / 0x80 * (camf.c6 + 1)
                self.c6_longitude_of_centre_of_hazard = self.a13_ellipse_centre_longitude + delta
            elif camf.a17 == 2:  # B3 – second ellipse definition
                self.a17_main_subject_for_specific_settings = \
                    qzss_dcx_camf_a17_main_subject_for_specific_settings[camf.a17]
                camf.c7 = self.extract_field(131, 2)
                camf.c8 = self.extract_field(133, 3)
                camf.c9 = self.extract_field(136, 5)
                camf.c10 = self.extract_field(141, 5)
                self.c7_shift_of_second_ellipse_centre = camf.c7
                self.c8_homothetic_factor_of_second_ellipse = 0.25 * (camf.c8 + 1)
                self.c9_bearing_angle_of_second_ellipse = camf.c9 * 360 / 32
                self.c10_guidance_library_for_second_ellipse_code = \
                    qzss_dcx_camf_c10_guidance_library_for_second_ellipse_code[camf.c10]
                self.c10_guidance_library_for_second_ellipse = \
                    qzss_dcx_camf_c10_guidance_library_for_second_ellipse[camf.c10]
            elif camf.a17 == 3:  # B4 – quantitative and detailed information related to hazard category and type
                self.a17_main_subject_for_specific_settings = \
                    qzss_dcx_camf_a17_main_subject_for_specific_settings[camf.a17]
                if camf.a4 == 36:  # earthquake
                    camf.d1 = self.extract_field(131, 4)
                    camf.d2 = self.extract_field(135, 3)
                    camf.d3 = self.extract_field(138, 4)
                    camf.d4 = self.extract_field(142, 4)
                    self.d1_magnitude_on_richter_scale = qzss_dcx_camf_d1_magnitude_on_richter_scale[camf.d1]
                    self.d2_seismic_coefficient = qzss_dcx_camf_d2_seismic_coefficient[camf.d2]
                    self.d3_azimuth_from_centre_of_main_ellipse_to_epicentre = \
                        qzss_dcx_camf_d3_azimuth_from_centre_of_main_ellipse_to_epicentre[camf.d3]
                    self.d4_vector_length_between_centre_of_main_ellipse_and_epicentre = \
                        qzss_dcx_camf_d4_vector_length_between_centre_of_main_ellipse_and_epicentre[camf.d4]
                elif camf.a4 == 44 or camf.a4 == 43:  # tsunami or tidal wave
                    camf.d5 = self.extract_field(131, 3)
                    self.d5_wave_height = qzss_dcx_camf_d5_wave_height[camf.d5]
                elif camf.a4 == 63 or camf.a4 == 71:  # cold wave or heat wave
                    camf.d6 = self.extract_field(131, 4)
                    self.d6_temperature_range = qzss_dcx_camf_d6_temperature_range[camf.d6]
                elif camf.a4 == 80:  # tropical cyclone (hurricane)
                    camf.d7 = self.extract_field(131, 3)
                    camf.d8 = self.extract_field(134, 4)
                    camf.d9 = self.extract_field(138, 3)
                    self.d7_hurricane_category = qzss_dcx_camf_d7_hurricane_category[camf.d7]
                    self.d8_wind_speed = qzss_dcx_camf_d8_wind_speed[camf.d8]
                    self.d9_rainfall_amounts = qzss_dcx_camf_d9_rainfall_amounts[camf.d9]
                elif camf.a4 == 82:  # tropical cyclone (typhoon)
                    camf.d36 = self.extract_field(131, 3)
                    camf.d8 = self.extract_field(134, 4)
                    camf.d9 = self.extract_field(138, 3)
                    self.d36_typhoon_category = qzss_dcx_camf_d36_typhoon_category[camf.d36]
                    self.d8_wind_speed = qzss_dcx_camf_d8_wind_speed[camf.d8]
                    self.d9_rainfall_amounts = qzss_dcx_camf_d9_rainfall_amounts[camf.d9]
                elif camf.a4 == 79:  # tornado
                    camf.d8 = self.extract_field(131, 4)
                    camf.d9 = self.extract_field(135, 3)
                    camf.d11 = self.extract_field(138, 3)
                    self.d8_wind_speed = qzss_dcx_camf_d8_wind_speed[camf.d8]
                    self.d9_rainfall_amounts = qzss_dcx_camf_d9_rainfall_amounts[camf.d9]
                    self.d11_tornado_probability = qzss_dcx_camf_d11_tornado_probability[camf.d11]
                elif camf.a4 == 77:  # storm or thunderstorm
                    camf.d8 = self.extract_field(131, 4)
                    camf.d9 = self.extract_field(135, 3)
                    camf.d10 = self.extract_field(138, 3)
                    camf.d16 = self.extract_field(141, 3)
                    self.d8_wind_speed = qzss_dcx_camf_d8_wind_speed[camf.d8]
                    self.d9_rainfall_amounts = qzss_dcx_camf_d9_rainfall_amounts[camf.d9]
                    self.d10_damage_category = qzss_dcx_camf_d10_damage_category[camf.d10]
                    self.d16_lightning_intensity = qzss_dcx_camf_d16_lightning_intensity[camf.d16]
                elif camf.a4 == 70:  # hail
                    camf.d12 = self.extract_field(131, 4)
                    self.d12_hail_scale = qzss_dcx_camf_d12_hail_scale[camf.d12]
                elif camf.a4 == 74:  # rainfall
                    camf.d9 = self.extract_field(131, 3)
                    camf.d13 = self.extract_field(134, 4)
                    self.d9_rainfall_amounts = qzss_dcx_camf_d9_rainfall_amounts[camf.d9]
                    self.d13_visibility = qzss_dcx_camf_d13_visibility[camf.d13]
                elif camf.a4 == 76:  # snowfall
                    camf.d14 = self.extract_field(131, 5)
                    camf.d13 = self.extract_field(136, 4)
                    self.d14_snow_depth = qzss_dcx_camf_d14_snow_depth[camf.d14]
                    self.d13_visibility = qzss_dcx_camf_d13_visibility[camf.d13]
                elif camf.a4 == 68:  # flood
                    camf.d15 = self.extract_field(131, 2)
                    self.d15_flood_severity = qzss_dcx_camf_d15_flood_severity[camf.d15]
                elif camf.a4 == 72:  # lightning
                    camf.d16 = self.extract_field(131, 3)
                    self.d16_lightning_intensity = qzss_dcx_camf_d16_lightning_intensity[camf.d16]
                elif camf.a4 == 81:  # wind chill / frost
                    camf.d8 = self.extract_field(131, 4)
                    camf.d6 = self.extract_field(135, 4)
                    self.d8_wind_speed = qzss_dcx_camf_d8_wind_speed[camf.d8]
                    self.d6_temperature_range = qzss_dcx_camf_d6_temperature_range[camf.d6]
                elif camf.a4 == 64:  # derecho
                    camf.d8 = self.extract_field(131, 4)
                    camf.d9 = self.extract_field(135, 3)
                    camf.d16 = self.extract_field(138, 3)
                    camf.d11 = self.extract_field(141, 3)
                    self.d8_wind_speed = qzss_dcx_camf_d8_wind_speed[camf.d8]
                    self.d9_rainfall_amounts = qzss_dcx_camf_d9_rainfall_amounts[camf.d9]
                    self.d16_lightning_intensity = qzss_dcx_camf_d16_lightning_intensity[camf.d16]
                    self.d11_tornado_probability = qzss_dcx_camf_d11_tornado_probability[camf.d11]
                elif camf.a4 == 69:  # fog
                    camf.d17 = self.extract_field(131, 3)
                    camf.d13 = self.extract_field(134, 4)
                    self.d17_fog_level = qzss_dcx_camf_d17_fog_level[camf.d17]
                    self.d13_visibility = qzss_dcx_camf_d13_visibility[camf.d13]
                elif camf.a4 == 75:  # snow storm / blizzard
                    camf.d13 = self.extract_field(131, 4)
                    camf.d8 = self.extract_field(134, 4)
                    self.d13_visibility = qzss_dcx_camf_d13_visibility[camf.d13]
                    self.d8_wind_speed = qzss_dcx_camf_d8_wind_speed[camf.d8]
                elif camf.a4 == 65:  # drought
                    camf.d18 = self.extract_field(131, 2)
                    self.d18_drought_level = qzss_dcx_camf_d18_drought_level[camf.d18]
                elif camf.a4 == 33:  # avalanche risk
                    camf.d19 = self.extract_field(131, 3)
                    self.d19_avalanche_warning_level = qzss_dcx_camf_d19_avalanche_warning_level[camf.d19]
                elif camf.a4 == 32:  # ash fall
                    camf.d20 = self.extract_field(131, 3)
                    self.d20_ash_fall_amount_and_impact = qzss_dcx_camf_d20_ash_fall_amount_and_impact[camf.d20]
                elif camf.a4 == 47:  # wind / wave / storm surge
                    camf.d8 = self.extract_field(131, 4)
                    camf.d5 = self.extract_field(135, 3)
                    self.d8_wind_speed = qzss_dcx_camf_d8_wind_speed[camf.d8]
                    self.d5_wave_height = qzss_dcx_camf_d5_wave_height[camf.d5]
                elif camf.a4 == 37:  # geomagnetic or solar storm
                    camf.d21 = self.extract_field(131, 3)
                    self.d21_geomagnetic_scale = qzss_dcx_camf_d21_geomagnetic_scale[camf.d21]
                elif camf.a4 == 103:  # terrorism
                    camf.d22 = self.extract_field(131, 3)
                    self.d22_terrorism_threat_level = qzss_dcx_camf_d22_terrorism_threat_level[camf.d22]
                elif camf.a4 == 27 or camf.a4 == 30:  # forest fire or risk of fire
                    camf.d23 = self.extract_field(131, 3)
                    self.d23_fire_risk_level = qzss_dcx_camf_d23_fire_risk_level[camf.d23]
                elif camf.a4 == 16 or camf.a4 == 18 or camf.a4 == 21:  # contaminated drinking water or marine pollution or river pollution
                    camf.d24 = self.extract_field(131, 3)
                    self.d24_water_quality = qzss_dcx_camf_d24_water_quality[camf.d24]
                elif camf.a4 == 23:  # uv radiation
                    camf.d25 = self.extract_field(131, 3)
                    self.d25_uv_index = qzss_dcx_camf_d25_uv_index[camf.d25]
                elif camf.a4 == 53 or camf.a4 == 51:  # risk of infection or pandemic
                    camf.d26 = self.extract_field(131, 5)
                    camf.d35 = self.extract_field(136, 6)
                    self.d26_number_of_cases_per_100000_inhabitants = \
                        qzss_dcx_camf_d26_number_of_cases_per_100000_inhabitants[camf.d26]
                    self.d35_infection_type = qzss_dcx_camf_d35_infection_type[camf.d35]
                elif camf.a4 == 19:  # noise pollution
                    camf.d27 = self.extract_field(131, 4)
                    self.d27_noise_range = qzss_dcx_camf_d27_noise_range[camf.d27]
                elif camf.a4 == 15:  # air pollution
                    camf.d28 = self.extract_field(131, 3)
                    self.d28_air_quality_index = qzss_dcx_camf_d28_air_quality_index[camf.d28]
                elif camf.a4 == 56 or camf.a4 == 57 or camf.a4 == 58 or camf.a4 == 55 or camf.a4 == 60:  # gas supply outage or outage of it system or power outage or emergency number outage or telephone line outage
                    camf.d29 = self.extract_field(131, 5)
                    self.d29_outage_estimated_duration = qzss_dcx_camf_d29_outage_estimated_duration[camf.d29]
                elif camf.a4 == 5:  # chemical hazard
                    camf.d31 = self.extract_field(131, 4)
                    self.d31_chemical_hazard_type = qzss_dcx_camf_d31_chemical_hazard_type[camf.d31]
                elif camf.a4 == 11 or camf.a4 == 9 or camf.a4 == 10:  # radiological hazard or nuclear hazard or nuclear power station accident
                    camf.d30 = self.extract_field(131, 4)
                    self.d30_nuclear_event_scale = qzss_dcx_camf_d30_nuclear_event_scale[camf.d30]
                elif camf.a4 == 4:  # biological hazard
                    camf.d32 = self.extract_field(131, 2)
                    camf.d33 = self.extract_field(133, 2)
                    self.d32_biohazard_level = qzss_dcx_camf_d32_biohazard_level[camf.d32]
                    self.d33_biohazard_type = qzss_dcx_camf_d33_biohazard_type[camf.d33]
                elif camf.a4 == 6:  # explosive hazard
                    camf.d34 = self.extract_field(131, 2)
                    self.d34_explosive_hazard_type = qzss_dcx_camf_d34_explosive_hazard_type[camf.d34]

        if self.ignore_ex1 is False:
            self.ex1_target_area = qzss_dcx_ex1_target_area_code_en[camf.ex1]
            self.ex1_target_area_ja = qzss_dcx_ex1_target_area_code_ja[camf.ex1]

        if self.ignore_ex2_to_ex7 is False:
            if camf.ex2 == 0:
                self.ex2_evacuate_direction_type = "Leave the additional target area range."
            else:
                self.ex2_evacuate_direction_type = "Head to the additional target area range."
            self.ex3_additional_ellipse_centre_latitude = -90 + 180 / 0x1FFFF * camf.ex3
            self.ex4_additional_ellipse_centre_longitude = 45 + 180 / 0x1FFFF * camf.ex4
            self.ex5_additional_ellipse_semi_major_axis = __get_axis(camf.ex5)
            self.ex6_additional_ellipse_semi_minor_axis = __get_axis(camf.ex6)
            self.ex7_additional_ellipse_azimuth = -90 + 180 / 0x80 * camf.ex7
        if self.ignore_ex8_to_ex9 is False:
            if camf.ex8 == 0:
                self.ex8_target_area_list_type = 'Prefecture code'
            else:
                self.ex8_target_area_list_type = 'Cities, towns and villages code'
            self.ex9_target_area_list = []
            self.ex9_target_area_list_ja = []
            if camf.ex8 == 0:  # prefecture code
                for key in qzss_dcx_camf_ex9_target_area_code_en.keys():
                    if key & camf.ex9 >> 17:
                        self.ex9_target_area_list.append(qzss_dcx_camf_ex9_target_area_code_en[key])
                        self.ex9_target_area_list_ja.append(qzss_dcx_camf_ex9_target_area_code_ja[key])
            else:  # cities, towns and villages code
                for i in range(4):
                    key = self.extract_field(147 + i * 16, 16)
                    if key != 0:
                        self.ex9_target_area_list.append(qzss_dcx_ex1_target_area_code_en[key])
                        self.ex9_target_area_list_ja.append(qzss_dcx_ex1_target_area_code_ja[key])

        self.dcx_version = camf.vn

        if dcx_message_type == DcxMessageType.OUTSIDE_JAPAN:
            return QzssDcxOutsideJapan(**self.get_params())
        elif dcx_message_type == DcxMessageType.L_ALERT:
            return QzssDcxLAlert(**self.get_params())
        elif dcx_message_type == DcxMessageType.J_ALERT:
            return QzssDcxJAlert(**self.get_params())
        elif dcx_message_type == DcxMessageType.LOCAL_GOV:
            return QzssDcxMTInfo(**self.get_params())
        else:  # has to be an unknown message
            return QzssDcxUnknown(**self.get_params())



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcr_jma_tsunami_height = QzssDcrDefinition(
    {
        1: "0.2m未満",
        2: "1m",
        3: "3m",
        4: "5m",
        5: "10m",
        6: "10m超",
        14: "不明",
        15: "その他の津波の高さ",
        # "NN*": "津波の高さ(コード番号：NN)",
    },
    undefined="津波の高さ(コード番号：%d)",
)



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcr_jma_tsunamigenic_potential_en = QzssDcrDefinition(
    {
        0: "There is No Possibility of a Tsunami",
        1: "There is a Possibility of a Destructive Ocean-Wide Tsunami",
        2: "There is a Possibility of a Destructive Regional Tsunami",
        3: "There is a Possibility of a Destructive Local Tsunami Near the Epicenter",
        4: "There is a Very Small Possibility of a destructive Local Tsunami",
        7: "There is Possibility of a Tsunami",  # "その他の津波発生の可能性有無",
        # "N*": "Tsunamigenic Potential (Code: N)",
    },
    undefined="Undefined Tsunamigenic Potential (Code: %d)"
)



from .qzss_dcr_decoder_jma_common import QzssDcrDecoderJmaCommon
from ..definition import qzss_dcr_jma_weather_forecast_region
from ..definition import qzss_dcr_jma_weather_related_disaster_sub_category
from ..definition import qzss_dcr_jma_weather_warning_state
from ..exception import QzssDcrDecoderException
from ..report import QzssDcReportJmaBase
from ..report import QzssDcReportJmaWeather


class QzssDcrDecoderJmaWeather(QzssDcrDecoderJmaCommon):
    schema = QzssDcReportJmaBase

    def decode(self):
        ar = self.extract_field(53, 3)
        try:
            self.weather_warning_state = qzss_dcr_jma_weather_warning_state[ar]
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined JMA Warning State: {ar}',
                self)
        self.weather_warning_state_raw = ar

        self.weather_related_disaster_sub_categories = []
        self.weather_related_disaster_sub_categories_raw = []
        self.weather_forecast_regions = []
        self.weather_forecast_regions_raw = []
        for i in range(6):
            offset = 56 + i * 24

            if self.extract_field(offset, 24) == 0:
                break

            ww = self.extract_field(offset, 5)
            try:
                self.weather_related_disaster_sub_categories.append(
                    qzss_dcr_jma_weather_related_disaster_sub_category[ww])
            except KeyError:
                raise QzssDcrDecoderException(
                    f'Undefined JMA Disaster Sub-Category: {ww}',
                    self)
            self.weather_related_disaster_sub_categories_raw.append(ww)

            pl = self.extract_field(offset + 5, 19)
            try:
                self.weather_forecast_regions.append(qzss_dcr_jma_weather_forecast_region[pl])
            except KeyError:
                raise QzssDcrDecoderException(
                    f'Undefined JMA Prefectural Forecast Region: {pl}',
                    self)
            self.weather_forecast_regions_raw.append(pl)

        return QzssDcReportJmaWeather(**self.get_params())



from .qzss_dcr_decoder_base import QzssDcrDecoderBase
from .qzss_dcr_decoder_jma import QzssDcrDecoderJma
from .qzss_dcx_decoder import QzssDcxDecoder
from ..definition import qzss_dcr_message_type
from ..definition import qzss_dcr_preamble
from ..exception import QzssDcrDecoderException
from ..report import QzssDcReportMessagePartial


class QzssDcrDecoder(QzssDcrDecoderBase):
    schema = QzssDcReportMessagePartial

    def decode(self):
        # checks the preamble
        pab = self.extract_field(0, 8)
        try:
            self.preamble = qzss_dcr_preamble[pab]
        except KeyError:
            raise QzssDcrDecoderException(
                f'Invalid Preamble: {pab}',
                self)

        # checks the crc
        crc = 0
        crc_remaining_len = 226
        data = self.message[:28] + bytes((self.message[28] & 0xC0,))  # clears the last 6 bits
        for byte in data:
            crc ^= (byte << 16)
            for _ in range(8):
                crc <<= 1
                if crc & 0x1000000:
                    crc ^= 0x1864cfb  # polynomial
                crc_remaining_len -= 1
                if crc_remaining_len == 0:
                    break
        crc &= 0xffffff
        if crc != self.extract_field(226, 24):
            raise QzssDcrDecoderException(
                'CRC Mismatch',
                self)

        # checks the message type
        mt = self.extract_field(8, 6)  # 6 bits
        try:
            self.message_type = qzss_dcr_message_type[mt]
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined Message Type: {mt}',
                self)

        if mt == 43:
            next_decoder = QzssDcrDecoderJma
        elif mt == 44:
            next_decoder = QzssDcxDecoder
        else:
            raise QzssDcrDecoderException(
                f'Unsupported Message Type: {mt}',
                self)

        # stacks the next decoder
        return next_decoder(**self.get_params()).decode()



from .qzss_dcr_decoder_jma_common import QzssDcrDecoderJmaCommon
from ..definition import qzss_dcr_jma_marine_forecast_region
from ..definition import qzss_dcr_jma_marine_warning_code
from ..exception import QzssDcrDecoderException
from ..report import QzssDcReportJmaBase
from ..report import QzssDcReportJmaMarine


class QzssDcrDecoderJmaMarine(QzssDcrDecoderJmaCommon):
    schema = QzssDcReportJmaBase

    def decode(self):
        self.marine_warning_codes = []
        self.marine_warning_codes_raw = []
        self.marine_forecast_regions = []
        self.marine_forecast_regions_raw = []
        for i in range(8):
            offset = 53 + i * 19
            dw = self.extract_field(offset, 5)
            pl = self.extract_field(offset + 5, 14)
            if dw == 0 and pl == 0:
                break

            try:
                self.marine_warning_codes.append(qzss_dcr_jma_marine_warning_code[dw])
            except KeyError:
                raise QzssDcrDecoderException(
                    f'Undefined JMA Marine Warning Code: {dw}',
                    self)
            self.marine_warning_codes_raw.append(dw)

            try:
                self.marine_forecast_regions.append(qzss_dcr_jma_marine_forecast_region[pl])
            except KeyError:
                raise QzssDcrDecoderException(
                    f'Undefined JMA Marine Forecast: {pl}',
                    self)
            self.marine_forecast_regions_raw.append(pl)

        return QzssDcReportJmaMarine(**self.get_params())



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcr_jma_coastal_region_en = QzssDcrDefinition(
    {
        1: "Ust-Kamchatsk (East Coasts of Kamchatka Peninsula)",
        2: "Petropavlovsk-K (East Coasts of Kamchatka Peninsula)",
        3: "Severo Kurilsk (Kuril Islands)",
        4: "Urup Islands (Kuril Islands)",
        5: "Busan (South Coasts of Korean Peninsula)",
        6: "Nohwa (South Coasts of Korean Peninsula)",
        7: "Seogwipo (South Coasts of Korean Peninsula)",
        8: "Hualien (Taiwan)",
        9: "Basco (East Coasts of Philippines)",
        10: "Palanan (East Coasts of Philippines)",
        11: "Legaspi (East Coasts of Philippines)",
        12: "Laoang (East Coasts of Philippines)",
        13: "Madrid (East Coasts of Philippines)",
        14: "Davao (East Coasts of Philippines)",
        15: "Berebere (North Coasts of Irian Jaya)",
        16: "Patani (North Coasts of Irian Jaya)",
        17: "Sorong (North Coasts of Irian Jaya)",
        18: "Manokwari (North Coasts of Irian Jaya)",
        19: "Warsa (North Coasts of Irian Jaya)",
        20: "Jayapura (North Coasts of Irian Jaya)",
        21: "Vanimo (North Coasts of Papua New Guinea)",
        22: "Wewak (North Coasts of Papua New Guinea)",
        23: "Madang (North Coasts of Papua New Guinea)",
        24: "Manus Islands (North Coasts of Papua New Guinea)",
        25: "Rabaul (North Coasts of Papua New Guinea)",
        26: "Kavieng (North Coasts of Papua New Guinea)",
        27: "Kimbe (North Coasts of Papua New Guinea)",
        28: "Kieta (North Coasts of Papua New Guinea)",
        29: "Guam (Mariana Islands)",
        30: "Saipan (Mariana Islands)",
        31: "Malakal (Palau)",
        32: "Yap Island (Micronesia)",
        33: "Chuuk Island (Micronesia)",
        34: "Pohnpei Island (Micronesia)",
        35: "Kosrae Island (Micronesia)",
        36: "Eniwetok Island (Marshall Islands)",
        37: "Panggoe (North Coasts of Solomon Islands)",
        38: "Auki (North Coasts of Solomon Islands)",
        39: "Kirakira (North Coasts of Solomon Islands)",
        40: "Munda (Solomon Sea)",
        41: "Honiara (Solomon Sea)",
        # 42:  "Reserved",
        # 43:  "Reserved",
        # 44:  "Reserved",
        # 45:  "Reserved",
        # 46:  "Reserved",
        # 47:  "Reserved",
        # 48:  "Reserved",
        # 49:  "Reserved",
        # 50:  "Reserved",
        # 51:  "Reserved",
        # 52:  "Reserved",
        # 53:  "Reserved",
        # 54:  "Reserved",
        # 55:  "Reserved",
        # 56:  "Reserved",
        # 57:  "Reserved",
        # 58:  "Reserved",
        # 59:  "Reserved",
        # 60:  "Reserved",
        # 61:  "Reserved",
        # 62:  "Reserved",
        # 63:  "Reserved",
        # 64:  "Reserved",
        # 65:  "Reserved",
        66: "Ostrov-Karaginskiy (East Coasts of Kamchatka Peninsula)",
        67: "Nikolskoya (East Coasts of Kamchatka Peninsula)",
        68: "Tongyeong (South Coasts of Korean Peninsula)",
        69: "Heuksando (South Coasts of Korean Peninsula)",
        70: "Cheju-Island (South Coasts of Korean Peninsula)",
        71: "Chilung (Taiwan)",
        72: "Taitung (Taiwan)",
        # 73:  "Reserved",
        74: "Homel (Taiwan)",
        75: "Geme (North Coasts of Irian Jaya)",
        76: "Ulamona (North Coasts of Papua New Guinea)",
        77: "Ghatere (North Coasts of Solomon Islands)",
        78: "Amun (Solomon Sea)",
        79: "Falamae (Solomon Sea)",
        80: "Misima (Solomon Sea)",
        81: "Alotau (Solomon Sea)",
        82: "Lae (Solomon Sea)",
        83: "Port-Moresby (Coral Sea)",
        84: "Shanghai (Coasts of East China Sea)",
        85: "Zhoushan (Coasts of East China Sea)",
        86: "Wenzhou (Coasts of East China Sea)",
        # 87:  "Reserved",
        # 88:  "Reserved",
        # 89:  "Reserved",
        # 90:  "Reserved",
        # 91:  "Reserved",
        # 92:  "Reserved",
        # 93:  "Reserved",
        # 94:  "Reserved",
        # 95:  "Reserved",
        # 96:  "Reserved",
        99: "Unknown",
        100: "Other region",
        # "NN*": "Coastal Region (Code: NN)",
    },
    undefined="Undefined Coastal Region (Code: %d)"
)



from .qzss_dcr_decoder_jma_common import QzssDcrDecoderJmaCommon
from ..definition import qzss_dcr_jma_eew_forecast_region
from ..definition import qzss_dcr_jma_long_period_ground_motion_lower_limit
from ..definition import qzss_dcr_jma_long_period_ground_motion_upper_limit
from ..definition import qzss_dcr_jma_seismic_intensity_lower_limit
from ..definition import qzss_dcr_jma_seismic_intensity_upper_limit
from ..exception import QzssDcrDecoderException
from ..report import QzssDcReportJmaBase
from ..report import QzssDcReportJmaEarthquakeEarlyWarning


class QzssDcrDecoderJmaEarthquakeEarlyWarning(QzssDcrDecoderJmaCommon):
    schema = QzssDcReportJmaBase

    def decode(self):
        lgll = self.extract_field(47, 3)
        try:
            self.long_period_ground_motion_lower_limit = qzss_dcr_jma_long_period_ground_motion_lower_limit[lgll]
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined JMA Expected Maximal Long-period Earthquake Ground Motion Lower Limit : {lgll}',
                self)
        self.long_period_ground_motion_lower_limit_raw = lgll 

        lgul = self.extract_field(50, 3)
        try:
            self.long_period_ground_motion_upper_limit = qzss_dcr_jma_long_period_ground_motion_upper_limit[lgul]
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined JMA Expected Maximal Long-period Earthquake Ground Motion Upper Limit : {lgul}',
                self)
        self.long_period_ground_motion_upper_limit_raw = lgul

        self.notifications_on_disaster_prevention, self.notifications_on_disaster_prevention_raw =\
            self.extract_notification_on_disaster_prevention_fields(53)
        self.occurrence_time_of_earthquake = self.extract_day_hour_min_field(80)
        self.depth_of_hypocenter, self.depth_of_hypocenter_raw = self.extract_depth_field(96)
        self.magnitude, self.magnitude_raw = self.extract_magnitude_field(105)
        self.seismic_epicenter, self.seismic_epicenter_raw = self.extract_seismic_epicenter_field(112)

        if self.depth_of_hypocenter_raw == 10 and self.magnitude_raw == 10:
            self.assumptive = True
        else:
            self.assumptive = False

        ll = self.extract_field(122, 4)
        try:
            self.seismic_intensity_lower_limit = qzss_dcr_jma_seismic_intensity_lower_limit[ll]
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined JMA Seismic Intensity Lower Limit : {ll}',
                self)
        self.seismic_intensity_lower_limit_raw = ll

        ul = self.extract_field(126, 4)
        try:
            self.seismic_intensity_upper_limit = qzss_dcr_jma_seismic_intensity_upper_limit[ul]
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined JMA Seismic Intensity Upper Limit : {ul}',
                self)
        self.seismic_intensity_upper_limit_raw = ul

        self.eew_forecast_regions = []
        self.eew_forecast_regions_raw = []
        for i in range(80):
            if self.extract_field(130 + i, 1) == 1:
                try:
                    self.eew_forecast_regions.append(qzss_dcr_jma_eew_forecast_region[i + 1])
                except KeyError:
                    raise QzssDcrDecoderException(
                        f'Undefined JMA EEW Forecast Region: {i + 1}',
                        self)
                self.eew_forecast_regions_raw.append(i + 1)

        return QzssDcReportJmaEarthquakeEarlyWarning(**self.get_params())



ublox_qzss_dcr_message_header: bytes = b'\xB5\x62\x02\x13'
# \xB5     -> ubx frame preamble sync char1
# \x62     -> ubx frame preamble sync char2
# \x02     -> message class: rbx
# \x13     -> message id: sfrbx



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcr_jma_information_serial_code = QzssDcrDefinition(
    {
        1: "調査中A（監視領域内でマグニチュード6.8以上の地震が発生したことにより、臨時に「南海トラフ沿いの地震に関する評価検討会」を開催）",
        2: "調査中B（1カ所以上のひずみ計での有意な変化と共に、他の複数の観測点でもそれに関係すると思われる変化が観測され、想定震源域内のプレート境界で通常と異なるゆっくりすべりりが発生している可能性がある場合など、ひずみ計で南海トラフ地震との関連性の検討が必要と認められる変化を観測したことにより、臨時に「南海トラフ沿いの地震に関する評価検討会」を開催）",
        3: "調査中C（その他、想定震源域内のプレート境界の固着状態の変化を示す可能性のある現象が観測される等、南海トラフ地震との関連性の検討が必要と認められる現象を観測したことにより、臨時に「南海トラフ沿いの地震に関する評価検討会」を開催）",
        4: "巨大地震警戒",
        5: "巨大地震注意",
        6: "調査終了",
        15: "その他の情報",
    },
    undefined="地震関連情報(コード番号：%d)"
)



from ..decoder import QzssDcrDecoder
from ..decoder import QzssDcrDecoderBase
from ..definition import nmea_qzss_dcr_message_header
from ..exception import QzssDcrDecoderException
from ..report import QzssDcReportBase


class NmeaQzssDcrDecoder(QzssDcrDecoderBase):
    schema = QzssDcReportBase

    def decode(self):
        if not self.sentence:
            raise EOFError('Encountered EOF')

        self.sentence = self.sentence.split()[0]

        if len(self.sentence) < 76:
            raise QzssDcrDecoderException(
                'Too Short Sentence',
                self)
        if len(self.sentence) > 76:
            raise QzssDcrDecoderException(
                'Too Long Sentence',
                self)

        # checks the checksum
        try:
            payload, csum = self.sentence.split('*')
        except ValueError:
            raise QzssDcrDecoderException(
                'Checksum Not Found',
                self)

        if len(csum) != 2:
            raise QzssDcrDecoderException(
                'Invalid Checksum Length',
                self)

        try:
            checksum = int(csum, 16)
        except ValueError:
            raise QzssDcrDecoderException(
                'Invalid Checksum',
                self)

        summed = 0
        for c in payload[1:]:  # without the '$' at the beginning
            summed ^= ord(c)

        if summed != checksum:
            raise QzssDcrDecoderException(
                'Checksum Mismatch, should be %02X' % summed,
                self)

        # extracts a message header, satellite id, and message
        try:
            self.message_header, sat_id, message_str = payload.split(',')
        except ValueError:
            raise QzssDcrDecoderException(
                'Invalid Sentence',
                self)

        # checks the message header
        if self.message_header != nmea_qzss_dcr_message_header:
            raise QzssDcrDecoderException(
                f'Unknown Message Header: {self.message_header}',
                self)

        # checks the satellite id
        if len(sat_id) != 2:
            raise QzssDcrDecoderException(
                f'Invalid Satellite ID: {sat_id}',
                self)
        self.satellite_id = int(sat_id)
        self.satellite_prn = self.satellite_id | 0x80

        # converts the message to bytes type
        try:
            self.message = bytes.fromhex(message_str + '0')  # padded with six 0s. the actual message size is 250 bits.
        except ValueError:
            raise QzssDcrDecoderException(
                'Invalid Message',
                self)

        # generates a nmea sentence
        self.nmea = self.message_to_nmea()

        # stacks the next decoder
        return QzssDcrDecoder(**self.get_params()).decode()



from .qzss_dcr_decoder_jma_common import QzssDcrDecoderJmaCommon
from ..definition import qzss_dcr_jma_flood_forecast_region
from ..definition import qzss_dcr_jma_flood_warning_level
from ..exception import QzssDcrDecoderException
from ..report import QzssDcReportJmaBase
from ..report import QzssDcReportJmaFlood


class QzssDcrDecoderJmaFlood(QzssDcrDecoderJmaCommon):
    schema = QzssDcReportJmaBase

    def decode(self):
        self.flood_warning_levels = []
        self.flood_warning_levels_raw = []
        self.flood_forecast_regions = []
        self.flood_forecast_regions_raw = []
        for i in range(3):
            offset = 53 + i * 44

            if self.extract_field(offset, 44) == 0:
                break

            lv = self.extract_field(offset, 4)
            try:
                self.flood_warning_levels.append(qzss_dcr_jma_flood_warning_level[lv])
            except KeyError:
                raise QzssDcrDecoderException(
                    f'Undefined JMA Flood Warning Level: {lv}',
                    self)
            self.flood_warning_levels_raw.append(lv)

            pl = self.extract_field(offset + 4, 40)
            try:
                self.flood_forecast_regions.append(qzss_dcr_jma_flood_forecast_region[pl])
            except KeyError:
                raise QzssDcrDecoderException(
                    f'Undefined JMA Flood Forecast Region: {pl}',
                    self)
            self.flood_forecast_regions_raw.append(pl)

        return QzssDcReportJmaFlood(**self.get_params())



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcx_camf_a9_selection_of_library = QzssDcrDefinition(
    {
        0: 'International library',
        1: 'Country/region guidance library',
    },
    undefined='Undefined Library (Code: %d)'
)



from datetime import datetime

from .qzss_dcr_decoder_base import QzssDcrDecoderBase
from .qzss_dcr_decoder_jma_ash_fall import QzssDcrDecoderJmaAshFall
from .qzss_dcr_decoder_jma_earthquake_early_warning import QzssDcrDecoderJmaEarthquakeEarlyWarning
from .qzss_dcr_decoder_jma_flood import QzssDcrDecoderJmaFlood
from .qzss_dcr_decoder_jma_hypocenter import QzssDcrDecoderJmaHypocenter
from .qzss_dcr_decoder_jma_marine import QzssDcrDecoderJmaMarine
from .qzss_dcr_decoder_jma_nankai_trough_earthquake import QzssDcrDecoderJmaNankaiTroughEarthquake
from .qzss_dcr_decoder_jma_northwest_pacific_tsunami import QzssDcrDecoderJmaNorthwestPacificTsunami
from .qzss_dcr_decoder_jma_seismic_intensity import QzssDcrDecoderJmaSeismicIntensity
from .qzss_dcr_decoder_jma_tsunami import QzssDcrDecoderJmaTsunami
from .qzss_dcr_decoder_jma_typhoon import QzssDcrDecoderJmaTyphoon
from .qzss_dcr_decoder_jma_volcano import QzssDcrDecoderJmaVolcano
from .qzss_dcr_decoder_jma_weather import QzssDcrDecoderJmaWeather
from ..definition import qzss_dcr_jma_report_classification
from ..definition import qzss_dcr_jma_report_classification_en
from ..definition import qzss_dcr_jma_disaster_category
from ..definition import qzss_dcr_jma_disaster_category_en
from ..definition import qzss_dcr_jma_information_type
from ..definition import qzss_dcr_jma_information_type_en
from ..exception import QzssDcrDecoderException
from ..report import QzssDcReportMessageBase


class QzssDcrDecoderJma(QzssDcrDecoderBase):
    schema = QzssDcReportMessageBase

    def decode(self):
        self.version = self.extract_field(214, 6)
        if self.version != 1:
            raise QzssDcrDecoderException(
                f'Unsupported JMA-DC Report Version: {self.version}',
                self)

        rc = self.extract_field(14, 3)
        try:
            self.report_classification = qzss_dcr_jma_report_classification[rc]
            self.report_classification_en = qzss_dcr_jma_report_classification_en[rc]
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined Report Classification: {rc}',
                self)
        self.report_classification_no = rc

        dc = self.extract_field(17, 4)
        try:
            self.disaster_category = qzss_dcr_jma_disaster_category[dc]
            self.disaster_category_en = qzss_dcr_jma_disaster_category_en[dc]
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined Disaster Category: {dc}',
                self)
        self.disaster_category_no = dc

        at_mo = self.extract_field(21, 4)
        if at_mo < 1 or at_mo > 12:
            raise QzssDcrDecoderException(
                f'Invalid Report Time: {at_mo} as month',
                self)
        at_d = self.extract_field(25, 5)
        if at_d < 1 or at_d > 31:
            raise QzssDcrDecoderException(
                f'Invalid Report Time: {at_d} as day',
                self)
        at_h = self.extract_field(30, 5)
        if at_h > 23:
            raise QzssDcrDecoderException(
                f'Invalid Report Time: {at_h} as hour',
                self)
        at_mi = self.extract_field(35, 6)
        if at_mi > 59:
            raise QzssDcrDecoderException(
                f'Invalid Report Time: {at_mi} as minute',
                self)

        at_y = self.timestamp.year
        if at_mo - self.timestamp.month > 6:
            at_y -= 1
        elif self.timestamp.month - at_mo > 6:
            at_y += 1

        if at_mo == 2 and at_d == 29:
            while at_y % 4 != 0 or (at_y % 100 == 0 and at_y % 400 != 0):
                at_y += 1

        self.report_time = datetime(year=at_y,
                                    month=at_mo,
                                    day=at_d,
                                    hour=at_h,
                                    minute=at_mi)

        it = self.extract_field(41, 2)
        try:
            self.information_type = qzss_dcr_jma_information_type[it]
            self.information_type_en = qzss_dcr_jma_information_type_en[it]
        except KeyError:
            raise QzssDcrDecoderException(
                'Undefined Information Type: {it}',
                self)
        self.information_type_no = it

        if dc == 1:
            next_decoder = QzssDcrDecoderJmaEarthquakeEarlyWarning
        elif dc == 2:
            next_decoder = QzssDcrDecoderJmaHypocenter
        elif dc == 3:
            next_decoder = QzssDcrDecoderJmaSeismicIntensity
        elif dc == 4:
            next_decoder = QzssDcrDecoderJmaNankaiTroughEarthquake
        elif dc == 5:
            next_decoder = QzssDcrDecoderJmaTsunami
        elif dc == 6:
            next_decoder = QzssDcrDecoderJmaNorthwestPacificTsunami
        elif dc == 8:
            next_decoder = QzssDcrDecoderJmaVolcano
        elif dc == 9:
            next_decoder = QzssDcrDecoderJmaAshFall
        elif dc == 10:
            next_decoder = QzssDcrDecoderJmaWeather
        elif dc == 11:
            next_decoder = QzssDcrDecoderJmaFlood
        elif dc == 12:
            next_decoder = QzssDcrDecoderJmaTyphoon
        elif dc == 14:
            next_decoder = QzssDcrDecoderJmaMarine
        else:
            raise QzssDcrDecoderException(
                f'Unsupported Disaster Category: {self.disaster_category}',
                self)

        return next_decoder(**self.get_params()).decode()



from .qzss_dcr_definition import QzssDcrDefinition


qzss_dcx_camf_a4_hazard_category = QzssDcrDefinition(
    {
        1: "CBRNE",
        2: "CBRNE",
        3: "CBRNE",
        4: "CBRNE",
        5: "CBRNE",
        6: "CBRNE",
        7: "CBRNE",
        8: "CBRNE",
        9: "CBRNE",
        10: "CBRNE",
        11: "CBRNE",
        12: "CBRNE",
        13: "CBRNE",
        14: "ENVIRONMENT",
        15: "ENVIRONMENT",
        16: "ENVIRONMENT",
        17: "ENVIRONMENT",
        18: "ENVIRONMENT",
        19: "ENVIRONMENT",
        20: "ENVIRONMENT",
        21: "ENVIRONMENT",
        22: "ENVIRONMENT",
        23: "ENVIRONMENT",
        24: "FIRE",
        25: "FIRE",
        26: "FIRE",
        27: "FIRE",
        28: "FIRE",
        29: "FIRE",
        30: "FIRE",
        31: "FIRE",
        32: "GEO",
        33: "GEO",
        34: "GEO",
        35: "GEO",
        36: "GEO",
        37: "GEO",
        38: "GEO",
        39: "GEO",
        40: "GEO",
        41: "GEO",
        42: "GEO",
        43: "GEO",
        44: "GEO",
        45: "GEO",
        46: "GEO",
        47: "GEO",
        48: "HEALTH",
        49: "HEALTH",
        50: "HEALTH",
        51: "HEALTH",
        52: "HEALTH",
        53: "HEALTH",
        54: "INFRASTRUCTURE",
        55: "INFRASTRUCTURE",
        56: "INFRASTRUCTURE",
        57: "INFRASTRUCTURE",
        58: "INFRASTRUCTURE",
        59: "INFRASTRUCTURE",
        60: "INFRASTRUCTURE",
        61: "MET",
        62: "MET",
        63: "MET",
        64: "MET",
        65: "MET",
        66: "MET",
        67: "MET",
        68: "MET",
        69: "MET",
        70: "MET",
        71: "MET",
        72: "MET",
        73: "MET",
        74: "MET",
        75: "MET",
        76: "MET",
        77: "MET",
        78: "MET",
        79: "MET",
        80: "MET",
        81: "MET",
        82: "MET",
        83: "RESCUE",
        84: "RESCUE",
        85: "RESCUE",
        86: "RESCUE",
        87: "RESCUE",
        88: "SAFETY",
        89: "SAFETY",
        90: "SAFETY",
        91: "SAFETY",
        92: "SAFETY",
        93: "SAFETY",
        94: "SAFETY",
        95: "SAFETY",
        96: "SAFETY",
        97: "SAFETY",
        98: "SECURITY",
        99: "SECURITY",
        100: "SECURITY",
        101: "SECURITY",
        102: "SECURITY",
        103: "SECURITY",
        104: "TRANSPORT",
        105: "TRANSPORT",
        106: "TRANSPORT",
        107: "TRANSPORT",
        108: "TRANSPORT",
        109: "TRANSPORT",
        110: "TRANSPORT",
        111: "TRANSPORT",
        112: "TRANSPORT",
        113: "OTHER",
    },
    undefined="UNDEFINED (Code: %d)"
)


qzss_dcx_camf_a4_hazard_type = QzssDcrDefinition(
    {
        1: "Air strike",
        2: "Attack on IT systems",
        3: "Attack with nuclear weapons",
        4: "Biological hazard",
        5: "Chemical hazard",
        6: "Explosive hazard",
        7: "Meteorite impact",
        8: "Missile attack",
        9: "Nuclear hazard",
        10: "Nuclear power station accident",
        11: "Radiological hazard",
        12: "Satellite/space re-entry debris",
        13: "Siren test",
        14: "Acid rain",
        15: "Air pollution",
        16: "Contaminated drinking water",
        17: "Gas leak",
        18: "Marine pollution",
        19: "Noise pollution",
        20: "Plague of insects",
        21: "River pollution",
        22: "Suspended dust",
        23: "UV radiation",
        24: "Conflagration",
        25: "Fire brigade deployment",
        26: "Fire gases",
        27: "Forest fire",
        28: "Fumes",
        29: "Odour nuisance",
        30: "Risk of fire",
        31: "Structure fire / Industrial fire",
        32: "Ash fall",
        33: "Avalanche risk",
        34: "Crack in the ground / sinkhole",
        35: "Debris flow",
        36: "Earthquake",
        37: "Geomagnetic or solar storm",
        38: "Glacial ice avalanche",
        39: "Landslide",
        40: "Lava flow",
        41: "Pyroclastic flow",
        42: "Snowdrifts",
        43: "Tidal wave",
        44: "Tsunami",
        45: "Volcanic mud flow",
        46: "Volcano eruption",
        47: "Wind / wave / storm surge",
        48: "Epizootic",
        49: "Food safety alert",
        50: "Health hazard",
        51: "Pandemic",
        52: "Pest infestation",
        53: "Risk of infection",
        54: "Building collapse",
        55: "Emergency number outage",
        56: "Gas supply outage",
        57: "Outage of IT systems",
        58: "Power outage",
        59: "Raw sewage",
        60: "Telephone line outage",
        61: "Black Ice",
        62: "Coastal flooding",
        63: "Cold wave",
        64: "Derecho",
        65: "Drought",
        66: "Dust storm",
        67: "Floating ice / icebergs",
        68: "Flood",
        69: "Fog",
        70: "Hail",
        71: "Heat wave",
        72: "Lightning",
        73: "Pollens",
        74: "Rainfall",
        75: "Snow storm / blizzard",
        76: "Snowfall",
        77: "Storm or thunderstorm",
        78: "Thawing",
        79: "Tornado",
        80: "Tropical cyclone (hurricane)",
        81: "Wind chill / frost",
        82: "Tropical cyclone (typhoon)",
        83: "Dam failure or bursting of a dam",
        84: "Dike failure or bursting of a dike",
        85: "Explosive ordnance disposal",
        86: "Factory accident",
        87: "Mine hazard",
        88: "Bomb / ammunition discovery",
        89: "Demonstration",
        90: "Hazardous material accident",
        91: "Life Threatening situation",
        92: "Major event",
        93: "Missing person / abduction",
        94: "Risk of explosion",
        95: "Safety warning",
        96: "Undefined flying object",
        97: "Unidentified animal",
        98: "Chemical attack",
        99: "Guerrilla attack",
        100: "Hijack",
        101: "Shooting or danger due to weapons",
        102: "Special forces attack",
        103: "Terrorism",
        104: "Aircraft crash",
        105: "Bridge collapse",
        106: "Dangerous goods accident",
        107: "Inland waterway transport accident",
        108: "Nautical disaster / Maritime / Marine Security",
        109: "Oil spill",
        110: "Road traffic incident",
        111: "Train/rail accident",
        112: "Tunnel accident",
        113: "Test alert",
    },
    undefined= "Undefined hazard type (Code: %d)",
)


qzss_dcx_camf_a4_hazard_definition = QzssDcrDefinition(
    {
        1: "Attack performed by an aircraft, usually of military type, either by bombing (or similar), firing guns or missiles.",
        2: "Intentional attempt to expose, alter, disable, destroy, steal or gain unauthorized access to or make unauthorised use of IT equipment.",
        3: "From www.dhs.gov 'Attack performed with a device producing a nuclear explosion.",
        4: "Presence of hazardous biological materials that present a risk or potential risk to the health of humans, animals, plants, or the environment.",
        5: "Presence of hazardous chemical materials that present a risk or potential risk to the health of humans, animals, plants, or the environment.",
        6: "Presence of hazardous explosive materials that present a risk or potential risk to the health of humans, animals, plants, or the environment.",
        7: "Meteorite entering from space into the earth atmosphere.",
        8: "Missile launch.",
        9: "Presence of hazardous nuclear materials that present a risk or potential risk to the health of humans, animals, plants, or the environment.",
        10: "Accident occurring at an installation where power is generated by a nuclear reactor.",
        11: "Presence of hazardous radioactive materials that present a risk or potential risk to the health of humans, animals, plants, or the environment.",
        12: "Satellite or parts of a satellite or launcher or space vehicle entering or re-entering from space into the earth atmosphere.",
        13: "Test of a warning device emitting a loud noise.",
        14: "Rain containing dissolved acidifying compounds, resulting from chemical pollution of the atmosphere by sulphur and nitrogen compounds. When deposited, these increase the acidity of the soil and water causing agricultural and ecological damage.",
        15: "Precise threshold is according to each local standard.",
        16: "Drinking water containing contaminants (germs and/or chemicals) above the quality standard. Precise threshold according to each local standard.",
        17: "Uncontrolled release of natural gas or gaseous product in the environment.",
        18: "Marine pollution occurs when substances used or spread by humans, such as industrial, agricultural and residential waste, particles, noise, excess carbon dioxide or invasive organisms enter the ocean and cause harmful effects there. Precise threshold is according to each local standard.",
        19: "Disturbing or unwanted noise that may present a potential risk to the health of humans, animals and wildlife.",
        20: "Infestation of insects or pervasive influx and development of insects or parasites affecting humans, animals, crops and materials.",
        21: "Contamination of water in rivers, usually as a result of human activities, in such a manner that negatively affects its legitimate uses.",
        22: "More general, particulates are microscopic solid particles suspended in the air. The source of the particulate matter can be natural or anthropogenic having adverse effect on human health. The IARC and WHO designate airborne particulates as a Group 1 carcinogen. Particulates are the most harmful form of air pollution due to their ability to penetrate deep into the lungs, blood streams and brain, causing health problems including heart attacks, respiratory disease, and premature death.",
        23: "Precise threshold is according to each local standard.",
        24: "A large and destructive fire that threatens human life, health, and/or property.",
        25: "Operation of firefighters in response to a fire or other emergency report. Note: Definition may vary, depending on the responsibility of the fire brigade in each country.",
        26: "Potentially dangerous gases emitted from a fire.",
        27: "A type of wildfire in a wooded area.",
        28: "An often-noxious suspension of particles in the air.",
        29: "Prolonged and offensive smell leading to the impairment of human well-being over a certain period of time at regular or irregular intervals.",
        30: "Increased likelihood of accidental fire.",
        31: "Fire affecting the structural parts of a building or a fire affecting an industrial site.",
        32: "Fine (less than 4 mm in diameter) unconsolidated volcanic debris blown into the atmosphere during an eruption; can remain airborne for long periods of time and travel considerable distance from the source.",
        33: "A large mass of loosened earth material, snow, or ice that slides, flows or falls rapidly down a mountainside under the force of gravity. Snow Avalanche: Rapid downslope movement of a mix of snow and ice. Debris Avalanche: The sudden and very rapid downslope movement of unsorted mass of rock and soil. There are two general types of debris avalanches - a cold debris avalanche usually results from an unstable slope suddenly collapsing whereas a hot debris avalanche results from volcanic activity leading to slope instability and collapse.",
        34: "Collapse of the land surface due to the dissolving of the subsurface rocks such as limestone or carbonate rock by water.",
        35: "Types of landslides that occur when heavy rain or rapid snow/ice melt send large amounts of vegetation, mud, or rock downslope by gravitational forces. Includes mud flow, rock fall.",
        36: "Sudden movement of a block of the Earth’s crust along a geological fault and associated ground shaking.",
        37: "A type of extra-terrestrial hazard caused by solar wind shockwaves that temporarily disturb the Earth’s magnetosphere. Geomagnetic storms can disrupt power grids, spacecraft operations, and satellite communications.",
        38: "Ice calving at end of a glacier.",
        39: "Independent of the presence of water, mass movement may also be triggered by earthquakes.",
        40: "The ejected magma that moves as a liquid mass downslope from a volcano during an eruption.",
        41: "Extremely hot gases, ash, and other materials of more than 1,000 ˚C that rapidly flow down the flank of a volcano (more than 700 km/h) during an eruption.",
        42: "Mass of snow heaped up by the wind and deposited along an obstruction or an irregularity of the terrain.",
        43: "Shallow water wave caused by the gravitational interactions between the Sun, Moon, and Earth. (from USGS)",
        44: "A series of waves (with long wavelengths when traveling across the deep ocean) that are generated by a displacement of massive amounts of water through underwater earthquakes, volcanic eruptions or landslides. Tsunami waves travel at very high speed across the ocean but as they begin to reach shallow water they slow down and the wave grows steeper.",
        45: "(Also known as Lahar) A mud flow resulting from a volcanic event. Hot or cold mixture of earthen material flowing on the slope of a volcano either during or between volcanic eruptions.",
        46: "A type of volcanic event near an opening/vent in the Earth’s surface including volcanic eruptions of lava, ash, hot vapour, gas, and pyroclastic material.",
        47: "An abnormal rise in sea level generated by a tropical cyclone or other intense storms.",
        48: "An epidemic outbreak of disease in an animal population, often with the implication that it may extend to humans.",
        49: "Presence of unsafe food containing harmful bacteria, viruses, parasites or chemical substances that present a risk to the health of humans, animals, plants or the environment.",
        50: "Potential source of danger for individual's health.",
        51: "An epidemic of an infectious disease that has spread across a large region, for instance multiple continents or worldwide, affecting a substantial number of people. A widespread endemic disease with a stable number of infected people is not a pandemic. Widespread endemic diseases with a stable number of infected people such as recurrences of seasonal influenza are generally excluded as they occur simultaneously in large regions of the globe rather than being spread worldwide.",
        52: "(from GEMET )1) The occurrence of one or more pest species in an area or location where their numbers and impact are currently or potentially at intolerable levels. 2) A sudden increase in destructiveness or population numbers of a pest species in a given area.",
        53: "Present of infectious agent(s) that present a risk or potential risk to the health of humans, animals, plants, or the environment.",
        54: "Sudden failure of the building structure resulting in its partial or total destruction.",
        55: "Interruption or failure of emergency number service.",
        56: "Interruption or failure of gas supply service.",
        57: "Interruption or failure of IT systems services.",
        58: "Interruption or failure of electricity distribution service.",
        59: "Release in the environment of untreated wastewater.",
        60: "Interruption or failure of telephone service.",
        61: "WMO definition: (1) Thin, new ice on freshwater or saltwater, appearing dark in colour because of its transparency, which is a result of its columnar grain structure. On lakes, black ice is commonly overlain by white ice formed from refrozen snow or slush. (2) A popular alternative for glaze. A thin sheet of ice, relatively dark in appearance, may form when light rain or drizzle falls on a road surface that is at a temperature below 0 °C or, alternatively, when water already on the road surface subsequently freezes when the temperature thereof falls below freezing point. It may also be formed when supercooled fog droplets are intercepted by buildings, fences and vegetation.",
        62: "Occurs when dry and low-lying land is submerged by seawater.",
        63: "Extreme cold and low temperatures.",
        64: "Widespread and usually fast-moving windstorms associated with convection/convective storm. Derechos include downburst and straight-line winds. The damage from derechos is often confused with the damage from tornadoes.",
        65: "An extended period of unusually low precipitation that produces a shortage of water for people, animals and plants. Drought is different from most other hazards in that it develops slowly, sometimes even over years, and its onset is generally difficult to detect.",
        66: "Strong winds carry particles of sand aloft, but generally confined to less than 50 feet (15 m), especially common in arid and semi-arid environments. A dust storm is also characterised by strong winds but carries smaller particles of dust rather than sand over an extensive area.",
        67: "Floating ice in rivers or icebergs in the open sea.",
        68: "A general term for the overflow of water from a stream channel onto normally dry land in the floodplain (riverine flooding), higher-than normal levels along the coast and in lakes or reservoirs (coastal flooding) as well as ponding of water at or near the point where the rain fell (flash floods).",
        69: "A suspension of very small, usually microscopic water droplets in the air, reducing visibility at the Earth’s surface.",
        70: "Solid precipitation in the form of irregular pellets or balls of ice more than 5 mm in diameter.",
        71: "A general term for temperature variations above (extreme heat).",
        72: "A high-voltage, visible electrical discharge produced by a thunderstorm and followed by the sound of thunder.",
        73: "Pollens that can cause human allergies are those of anemophilous plants (pollen is dispersed by air currents.) Such plants produce large quantities of lightweight pollen (because wind dispersal is random and the likelihood of one pollen grain landing on another flower is small), which can be carried for great distances and are easily inhaled by humans.",
        74: "Rainfall greater than or equal to 50 mm in past 24 hours. Note: Precise threshold is according to each local standard.",
        75: "A low-pressure system in winter months with significant accumulations of snow, freezing rain, sleet or ice. A blizzard is a severe snowstorm with winds exceeding 35 mph (56 km/h) for three or more hours, producing reduced visibility (less than .25 mile (400 m).",
        76: "Snowfall greater than or equal to 50 mm in past 24 hours. Note: Precise threshold is according to each local standard.",
        77: "Meteorological event generating winds equal or higher than 10 on the Beaufort scale.",
        78: "Gradual warning causing ground ice to degrade, potentially leading the soil surface to collapse.",
        79: "A violently rotating column of air that reaches the ground or open water (waterspout).",
        80: "A tropical cyclone originates over tropical or subtropical waters. It is characterised by a warm-core, non-frontal synoptic-scale cyclone with a low-pressure centre, spiral rain bands and strong winds. Depending on their location, tropical cyclones are referred to as hurricanes (Atlantic, Northeast Pacific), typhoons (Northwest Pacific), or cyclones (South Pacific and Indian Ocean). With a minimum wind speed of 119 km/h (or 74 miles/h) a tropical storm in the Atlantic or Northeast Pacific is called a hurricane.",
        81: "Frost is the consequence of radiative cooling resulting in the formation of thin ice crystals on the ground or other surfaces in the form of needles, feathers, scales, or fans. Frost occurs when the temperature of surfaces is below freezing and water vapor from humid air forms solid deposits on the cold surface. Freeze occurs when the air temperature is at (32 °F / 0 °C) or below over a widespread area for a climatologically significant period of time. Use of the term is usually restricted to advective situations or to occasions when wind or other conditions prevent frost. Frost and freeze are particularly damaging during the crop growing season.",
        82: "A tropical cyclone originates over tropical or subtropical waters. It is characterised by a warm-core, non-frontal synoptic-scale cyclone with a low-pressure centre, spiral rain bands and strong winds. Depending on their location, tropical cyclones are referred to as hurricanes (Atlantic, Northeast Pacific), typhoons (Northwest Pacific), or cyclones (South Pacific and Indian Ocean). With a minimum wind speed of 119 km/h (or 74 miles/h) a tropical storm in the Atlantic or Northeast Pacific is called a hurricane.",
        83: "Incident at a dam potentially leading to the uncontrolled release of the dam content.",
        84: "Incident at a dike potentially leading to the uncontrolled release of water.",
        85: "Activity to render safe all types of ordnance, including explosive, chemical and biological. It covers land and underwater location, identification, render-safe, and recovery (or disposal) of ordnance.",
        86: "Incident at an industrial site affecting the surrounding area.",
        87: "Any of the dangers peculiar to the winning and working of coal and minerals. These include collapse of ground, explosion of released gas, inundation by water, spontaneous combustion, inhalation of dust and poisonous gases, etc.",
        88: "Unexpected presence of bomb or ammunition.",
        89: "A march or gathering of people taking part in to show their opposition to something or their support for something.",
        90: "An unexpected uncontrolled release of substances that have the potential to harm a person or the environment upon contact.",
        91: "Emergency situation that requires immediate intervention to save the life of the persons concerned.",
        92: "Planned happening which may generate significant disruption in road traffic or transport.",
        93: "Unaccounted person whose health status (alive/dead) cannot be confirmed.",
        94: "Intentional use of explosive materials aiming at killing or damaging the health of humans, animals, plants, or the environment.",
        95: "Announcement of a potentially dangerous situation or event to the public impacting their safety.",
        96: "Presence of an unidentified flying object which may present a risk or potential risk to the health of humans, animals, plants, or the environment.",
        97: "Presence of an unidentified animal which may present a risk or a potential risk to the health of humans, animals, plants, or the environment.",
        98: "Intentional use of dangerous chemical materials aiming at killing or damaging the health of humans, animals, plants, or the environment.",
        99: "Paramilitary armed forces carrying violent actions.",
        100: "Airplane or other vehicle under the control of hijackers.",
        101: "Dangerous use of firearm(s) or weapons in a populated area.",
        102: "Military armed forces carrying violent actions.",
        103: "Terrorist action aiming at generating a high number of victims or major material damage.",
        104: "Aircraft crash.",
        105: "Structural failure of a bridge resulting in its partial or total destruction.",
        106: "Accident occurrence associated with and related to the conveyance of dangerous goods by air, road, inland waterways and sea, which results in fatal or serious injury to a person or major damage to property or environment.",
        107: "Ship accident on an inland waterway.",
        108: "Sinking ship or any other event occurring at sea which may present of security or safety risk.",
        109: "Leak of oil or pollutant from a ship or other transportation mean resulting in pollution.",
        110: "Announcement of road traffic disruption.",
        111: "Accident occurring on a railway and involving train(s) and/or other vehicles.",
        112: "Car crash or railway accident occurring in a tunnel.",
        113: "Test alert message for end-to-end testing.",
    },
    undefined= "Undefined hazard definition. (Code: %d)",
)



from .qzss_dcr_definition import QzssDcrDefinition


qzss_dcx_camf_a11_international_library_code = QzssDcrDefinition(
    {
        0: "IC-A-01",
        1: "IC-A-02",
        2: "IC-A-03",
        3: "IC-A-04",
        4: "IC-A-05",
        5: "IC-A-06",
        6: "IC-A-07",
        7: "IC-A-08",
        8: "IC-A-09",
        9: "IC-A-10",
        10: "IC-A-11",
        11: "IC-A-12",
        12: "IC-A-13",
        13: "IC-A-14",
        14: "IC-A-15",
        15: "IC-A-16",
        16: "IC-A-17",
        17: "IC-A-18",
        18: "IC-A-19",
        19: "IC-A-20",
        20: "IC-A-21",
        21: "IC-A-22",
        22: "IC-A-23",
        23: "IC-A-24",
        24: "IC-A-25",
        25: "IC-A-26",
        26: "IC-A-27",
        27: "IC-A-28",
        28: "IC-A-29",
        29: "IC-A-30",
        30: "IC-A-31",
        31: "IC-A-32",
    },
    undefined='IC-A-UNDEFINED (Code: %d)'
)


qzss_dcx_camf_a11_international_library = QzssDcrDefinition(
    {
        0: "",
        1: "You are in the danger zone, leave the area immediately. Listen to radio or media for directions and information.",
        2: "You are in the danger zone, leave the area immediately and reach the evacuation point indicated by the area plotted in yellow. Listen to radio or media for directions and information.",
        3: "Seek shelter in a building immediately. Stay under cover and stay informed.",
        4: "Seek out a cellar or interior rooms on lower floors.",
        5: "If you are in an alpine terrain, start descending immediately and seek for shelter.",
        6: "Quickly move into interior rooms. If you are in a vehicle: Stop driving immediately on the edge of the road. If a building is nearby, seek shelter in that building.",
        7: "If you are in open terrain and you cannot find shelter, lie face-down on the ground and protect your head and neck with your hands, in a hollow where possible.",
        8: "Prepare for evacuation. Take only the essentials with you, especially ID cards, passport, credit cards and cash. Evacuate only after the instruction of the emergency authorities.",
        9: "Prepare emergency food and relief material: Check and restock your equipment and supplies of water, food, medicine, cash and batteries.",
        10: "Stay away from glass surfaces such as windows and glass doors. There is a risk of injury from glass splinters.",
        11: "Reduce your power consumption to a minimum.",
        12: "Reduce your water consumption to a minimum.",
        13: "Boil water before drinking it or using it in the kitchen.",
        14: "Keep at least one metre away from any conversation partners. Avoid physical contact with other people such as kissing and shaking hands. Wash your hands regularly and thoroughly.",
        15: "Do not drink any tap water. Avoid any skin contact with tap water. Only drink mineral water from a bottle. Turn off the water supply to your house.",
        16: "Watch out for escaping gas. This can be indicated by hissing noises or a typical gas odour. Do not use matches, lighters or the like: naked flames in combination with leaking gas can lead to explosions and fires.",
        17: "Do not go outside and do not use your car.",
        18: "Do not touch any objects that seem suspicious to you. Debris can cause additional hazards such as fires and explosions. Inform the emergency services about damage and debris.",
        19: "Do not enter smoke-filled rooms. Deadly gases can form there.",
        20: "Do not enter cellars or underground car parks.",
        21: "Do not leave pets or livestock outside.",
        22: "Do not touch any dead animals. Report any findings of dead wild animals to the authorities.",
        23: "Avoid driving.",
        24: "Avoid all items with metal parts such as umbrellas and bicycles. Do not bathe or shower during a thunderstorm. Bathing and showering can be life-threatening.",
        25: "Avoid rooms directly underneath the roof truss. Avoid very large rooms, such as halls, in which the ceiling is not supported by pillars.",
        26: "Avoid going outdoors. Keep away from trees, towers and masts. Keep at least 20 m away from power lines. Watch out for flying objects and falling objects.",
        27: "Avoid the danger area.",
        28: "Avoid going out when it is not necessary.",
        29: "This is only a test. You do not have to take any action or to adopt any particular sheltering behaviour.",
        30: "This replaces the warning previously in effect for this area.",
        31: "Conditions have improved and are no longer expected to meet alert criteria.",
    },
    undefined='Undefined guidance to react instruction in the international library. (Code: %d)'
)



from .qzss_dcr_decoder_jma_common import QzssDcrDecoderJmaCommon
from ..definition import qzss_dcr_jma_volcanic_warning_code
from ..definition import qzss_dcr_jma_volcano_name
from ..exception import QzssDcrDecoderException
from ..report import QzssDcReportJmaBase
from ..report import QzssDcReportJmaVolcano


class QzssDcrDecoderJmaVolcano(QzssDcrDecoderJmaCommon):
    schema = QzssDcReportJmaBase

    def decode(self):
        self.ambiguity_of_activity_time_no = self.extract_field(50, 3)
        self.activity_time = self.extract_day_hour_min_field(53)

        dw = self.extract_field(69, 7)
        try:
            self.volcanic_warning_code = qzss_dcr_jma_volcanic_warning_code[dw]
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined JMA Volcanic Warning Code: {dw}',
                self)
        self.volcanic_warning_code_raw = dw

        vo = self.extract_field(76, 12)
        try:
            self.volcano_name = qzss_dcr_jma_volcano_name[vo]
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined JMA Volcano Name: {vo}',
                self)
        self.volcano_name_raw = vo

        self.local_governments = []
        self.local_governments_raw = []
        for i in range(5):
            offset = 88 + i * 23
            if self.extract_field(offset, 23) == 0:
                break
            local_government, lg = self.extract_local_government(offset)
            self.local_governments.append(local_government)
            self.local_governments_raw.append(lg)

        return QzssDcReportJmaVolcano(**self.get_params())



import struct

from ..definition import qzss_dcr_message_type
from ..definition import ublox_qzss_dcr_message_header

buffers = {}


def __pop(size, reader, reader_args=None, reader_kwargs=None):
    buf = buffers.get(reader)
    if buf is None:
        buf = b''

    if reader_args is None:
        reader_args = ()

    if reader_kwargs is None:
        reader_kwargs = {}

    while size > len(buf):
        data = reader(*reader_args, **reader_kwargs)
        if not data:
            raise EOFError('Encountered EOF')
        buf += data

    ret = buf[:size]
    buffers.update({reader: buf[size:]})

    return ret


def ublox_qzss_dcr_message_extractor(reader, reader_args=None, reader_kwargs=None):

    if reader_args is None:
        reader_args = ()

    if reader_kwargs is None:
        reader_kwargs = {}

    match_count = 0
    while True:
        if ublox_qzss_dcr_message_header[match_count] == __pop(1,
                                                               reader,
                                                               reader_args,
                                                               reader_kwargs)[0]:
            match_count += 1
        else:
            match_count = 0

        if match_count == len(ublox_qzss_dcr_message_header): # SFRBX message
            match_count = 0
            message_length_bytes = __pop(2,
                                         reader,
                                         reader_args,
                                         reader_kwargs)
            message_length = struct.unpack('<H', message_length_bytes)[0]
            message = ublox_qzss_dcr_message_header + \
                      message_length_bytes + \
                      __pop(message_length + 2 , # payload + CK_A + CK_B
                            reader,
                            reader_args,
                            reader_kwargs)
            
            if message[6] != 5:  # not sent by QZSS
                continue

            if message[8] != 1:  # not a L1S signal
                continue

            if message[16] >> 2 not in qzss_dcr_message_type.keys():
                continue

            return message



from .qzss_dcr_definition import QzssDcrDefinition

qzss_dcx_camf_a2_country_region_name = QzssDcrDefinition(
    {
        0: "Afghanistan",
        1: "Albania",
        2: "Antarctica",
        3: "Algeria",
        4: "American Samoa",
        5: "Andorra",
        6: "Angola",
        7: "Antigua and Barbuda",
        8: "Azerbaijan",
        9: "Argentina",
        10: "Australia",
        11: "Austria",
        12: "The Bahamas",
        13: "Bahrain",
        14: "Bangladesh",
        15: "Armenia",
        16: "Barbados",
        17: "Belgium",
        18: "Bermuda",
        19: "Bhutan",
        20: "Plurinational State of Bolivia ",
        21: "Bosnia and Herzegovina",
        22: "Botswana",
        23: "Bouvet Island",
        24: "Brazil",
        25: "Belize",
        26: "The British Indian Ocean Territory",
        27: "Solomon Islands",
        28: "British Virgin Islands",
        29: "Brunei Darussalam",
        30: "Bulgaria",
        31: "Myanmar",
        32: "Burundi",
        33: "Belarus",
        34: "Cambodia",
        35: "Cameroon",
        36: "Canada",
        37: "Cabo Verde",
        38: "The Cayman Islands",
        39: "The Central African Republic",
        40: "Sri Lanka",
        41: "Chad",
        42: "Chile",
        43: "China",
        44: "Taiwan",
        45: "Christmas Island",
        46: "The Cocos (Keeling) Islands",
        47: "Colombia",
        48: "The Comoros the",
        49: "Mayotte",
        50: "The Congo",
        51: "The Democratic Republic of the Congo",
        52: "The Cook Islands",
        53: "Costa Rica",
        54: "Croatia",
        55: "Cuba",
        56: "Cyprus",
        57: "Czechia",
        58: "Benin",
        59: "Denmark",
        60: "Dominica",
        61: "The Dominican Republic",
        62: "Ecuador",
        63: "El Salvador",
        64: "Equatorial Guinea",
        65: "Ethiopia",
        66: "Eritrea",
        67: "Estonia",
        68: "The Faroe Islands",
        69: "The Falkland Islands [Malvinas]",
        70: "South Georgia and the South Sandwich Islands",
        71: "Fiji",
        72: "Finland",
        73: "Åland Islands",
        74: "France",
        75: "French Guiana",
        76: "French Polynesia",
        77: "The French Southern Territories",
        78: "Djibouti",
        79: "Gabon",
        80: "Georgia",
        81: "The Gambia",
        82: "State of Palestine",
        83: "Germany",
        84: "Ghana",
        85: "Gibraltar",
        86: "Kiribati",
        87: "Greece",
        88: "Greenland",
        89: "Grenada",
        90: "Guadeloupe",
        91: "Guam",
        92: "Guatemala",
        93: "Guinea",
        94: "Guyana",
        95: "Haiti",
        96: "Heard Island and McDonald Islands",
        97: "The Holy See",
        98: "Honduras",
        99: "Hong Kong",
        100: "Hungary",
        101: "Iceland",
        102: "India",
        103: "Indonesia",
        104: "Islamic Republic of Iran",
        105: "Iraq",
        106: "Ireland",
        107: "Israel",
        108: "Italy",
        109: "Côte d'Ivoire",
        110: "Jamaica",
        111: "Japan",
        112: "Kazakhstan",
        113: "Jordan",
        114: "Kenya",
        115: "The Democratic People's Republic of Korea",
        116: "The Republic of Korea",
        117: "Kuwait",
        118: "Kyrgyzstan",
        119: "The Lao People's Democratic Republic",
        120: "Lebanon",
        121: "Lesotho",
        122: "Latvia",
        123: "Liberia",
        124: "Libya",
        125: "Liechtenstein",
        126: "Lithuania",
        127: "Luxembourg",
        128: "Macao",
        129: "Madagascar",
        130: "Malawi",
        131: "Malaysia",
        132: "Maldives",
        133: "Mali",
        134: "Malta",
        135: "Martinique",
        136: "Mauritania",
        137: "Mauritius",
        138: "Mexico",
        139: "Monaco",
        140: "Mongolia",
        141: "The Republic of Moldova",
        142: "Montenegro",
        143: "Montserrat",
        144: "Morocco",
        145: "Mozambique",
        146: "Oman",
        147: "Namibia",
        148: "Nauru",
        149: "Nepal",
        150: "The Netherlands",
        151: "Curaçao",
        152: "Aruba",
        153: "Sint Maarten (Dutch part)",
        154: "Bonaire, Sint Eustatius and Saba",
        155: "New Caledonia",
        156: "Vanuatu",
        157: "New Zealand",
        158: "Nicaragua",
        159: "The Niger",
        160: "Nigeria",
        161: "Niue",
        162: "Norfolk Island",
        163: "Norway",
        164: "The Northern Mariana Islands",
        165: "The United States Minor Outlying Islands",
        166: "Federated States of Micronesia",
        167: "The Marshall Islands",
        168: "Palau",
        169: "Pakistan",
        170: "Panama",
        171: "Papua New Guinea",
        172: "Paraguay",
        173: "Peru",
        174: "The Philippines",
        175: "Pitcairn",
        176: "Poland",
        177: "Portugal",
        178: "Guinea-Bissau",
        179: "Timor-Leste",
        180: "Puerto Rico",
        181: "Qatar",
        182: "Réunion",
        183: "Romania",
        184: "The Russian Federation",
        185: "Rwanda",
        186: "Saint Barthélemy",
        187: "Saint Helena, Ascension and Tristan da Cunha",
        188: "Saint Kitts and Nevis",
        189: "Anguilla",
        190: "Saint Lucia",
        191: "Saint Martin (French part)",
        192: "Saint Pierre and Miquelon",
        193: "Saint Vincent and the Grenadines",
        194: "San Marino",
        195: "Sao Tome and Principe",
        196: "Saudi Arabia",
        197: "Senegal",
        198: "Serbia",
        199: "Seychelles",
        200: "Sierra Leone",
        201: "Singapore",
        202: "Slovakia",
        203: "Viet Nam",
        204: "Slovenia",
        205: "Somalia",
        206: "South Africa",
        207: "Zimbabwe",
        208: "Spain",
        209: "South Sudan",
        210: "The Sudan",
        211: "Western Sahara*",
        212: "Suriname",
        213: "Svalbard and Jan Mayen",
        214: "Eswatini",
        215: "Sweden",
        216: "Switzerland",
        217: "The Syrian Arab Republic",
        218: "Tajikistan",
        219: "Thailand",
        220: "Togo",
        221: "Tokelau",
        222: "Tonga",
        223: "Trinidad and Tobago",
        224: "The United Arab Emirates",
        225: "Tunisia",
        226: "Turkey",
        227: "Turkmenistan",
        228: "The Turks and Caicos Islands",
        229: "Tuvalu",
        230: "Uganda",
        231: "Ukraine",
        232: "North Macedonia",
        233: "Egypt",
        234: "The United Kingdom of Great Britain and Northern Ireland",
        235: "Guernsey",
        236: "Jersey",
        237: "Isle of Man",
        238: "Tanzania, the United Republic of",
        239: "The United States of America",
        240: "Virgin Islands (U.S.)",
        241: "Burkina Faso",
        242: "Uruguay",
        243: "Uzbekistan",
        244: "Bolivarian Republic of Venezuela",
        245: "Wallis and Futuna",
        246: "Samoa",
        247: "Yemen",
        248: "Zambia",
        500: "EU Organisations",
        501: "UN Organisations",
        502: "International",
    },
    undefined = "Undefined Country/Region (Code: %d)"
)



from .qzss_dcr_decoder_jma_common import QzssDcrDecoderJmaCommon
from ..report import QzssDcReportJmaBase
from ..report import QzssDcReportJmaHypocenter


class QzssDcrDecoderJmaHypocenter(QzssDcrDecoderJmaCommon):
    schema = QzssDcReportJmaBase

    def decode(self):
        self.notifications_on_disaster_prevention, self.notifications_on_disaster_prevention_raw =\
            self.extract_notification_on_disaster_prevention_fields(53)
        self.occurrence_time_of_earthquake = self.extract_day_hour_min_field(80)
        self.depth_of_hypocenter, self.depth_of_hypocenter_raw = self.extract_depth_field(96)
        self.magnitude, self.magnitude_raw = self.extract_magnitude_field(105)
        self.seismic_epicenter, self.seismic_epicenter_raw = self.extract_seismic_epicenter_field(112)
        self.coordinates_of_hypocenter = self.extract_lat_lon_field(122)
        return QzssDcReportJmaHypocenter(**self.get_params())



from .qzss_dcr_decoder_jma_common import QzssDcrDecoderJmaCommon
from ..definition import qzss_dcr_jma_coastal_region_en
from ..definition import qzss_dcr_jma_northwest_pacific_tsunami_height_en
from ..definition import qzss_dcr_jma_tsunamigenic_potential_en
from ..exception import QzssDcrDecoderException
from ..report import QzssDcReportJmaBase
from ..report import QzssDcReportJmaNorthwestPacificTsunami


class QzssDcrDecoderJmaNorthwestPacificTsunami(QzssDcrDecoderJmaCommon):
    schema = QzssDcReportJmaBase

    def decode(self):
        tp = self.extract_field(53, 3)
        try:
            self.tsunamigenic_potential_en = qzss_dcr_jma_tsunamigenic_potential_en[tp]
        except KeyError:
            raise QzssDcrDecoderException(
                f'Undefined JMA Tsunamigenic Potential : {tp}',
                self)
        self.tsunamigenic_potential_raw = tp

        self.expected_tsunami_arrival_times = []
        self.tsunami_heights_en = []
        self.tsunami_heights_raw = []
        self.coastal_regions_en = []
        self.coastal_regions_raw = []
        for i in range(5):
            offset = 56 + i * 28

            if self.extract_field(offset, 28) == 0:
                break

            self.expected_tsunami_arrival_times.append(self.extract_expected_tsunami_arrival_time(offset))

            th = self.extract_field(offset + 12, 9)
            try:
                self.tsunami_heights_en.append(qzss_dcr_jma_northwest_pacific_tsunami_height_en[th])
            except KeyError:
                raise QzssDcrDecoderException(
                    f'Undefined JMA Northwest Pacific Tsunami Height: {th}',
                    self)
            self.tsunami_heights_raw.append(th)

            pl = self.extract_field(offset + 21, 7)
            try:
                self.coastal_regions_en.append(qzss_dcr_jma_coastal_region_en[pl])
            except KeyError:
                raise QzssDcrDecoderException(
                    f'Undefined JMA Coastal Region: {pl}',
                    self)
            self.coastal_regions_raw.append(pl)

        return QzssDcReportJmaNorthwestPacificTsunami(**self.get_params())



import warnings


class QzssDcrDefinition(dict):
    def __init__(self,
                 default,
                 prefix=None,
                 prefix_extractor=None,
                 undefined=None,
                 *args,
                 **kwargs):
        super().__init__(default, *args, **kwargs)

        if (prefix is None) != (prefix_extractor is None):
            warnings.warn(
                "prefix and prefix_extractor should be both set or both None",
                RuntimeWarning,
                stacklevel=2
            )

        if undefined is not None and not isinstance(undefined, str) and prefix is not None:
            warnings.warn(
                "prefix is ignored when undefined is not a string",
                RuntimeWarning,
                stacklevel=2
            )

        self.prefix = prefix
        self.prefix_extractor = prefix_extractor
        self.undefined = undefined

    def __missing__(self, key):
        # for non-string undefined values, return as-is
        if self.undefined is not None and not isinstance(self.undefined, str):
            return self.undefined

        # for string values, apply formatting
        value = None

        if self.prefix and self.prefix_extractor:
            value = self.prefix.get(self.prefix_extractor(key))
            if value is not None:
                value = value % key

        if value is None and self.undefined is not None:
            value = self.undefined % key

        if value is None:
            raise KeyError(key)

        return value



from .qzss_dcr_definition import QzssDcrDefinition

ublox_qzss_svid_prn_map : QzssDcrDefinition = QzssDcrDefinition(
    {
        2: 184,
        3: 185,
        4: 186,
        7: 189,
    },
    undefined = 183 # has already gone silent, but it is still used as the default value
)



from ..definition import nmea_qzss_dcr_message_header
from ..exception import QzssDcrDecoderNotImplementedError


class QzssDcrDecoderBase:
    schema = None

    def __init__(self, sentence, **kwargs):
        self.set_params(sentence=sentence, **kwargs)

    def decode(self):
        raise QzssDcrDecoderNotImplementedError('Decoder Not Implemented')

    def message_to_nmea(self):
        sat_id = getattr(self, 'satellite_id', None)
        if sat_id is None:
            sat_id = 55  # Set the satellite_id of PRN183 if it was default.

        nmea_partial = '{header},{satellite_id},{message}'.format(header=nmea_qzss_dcr_message_header,
                                                                  satellite_id=sat_id,
                                                                  message=self.message.hex()[:-1].upper())

        checksum = 0
        for c in nmea_partial[1:]:  # without the '$' at the beginning
            checksum ^= ord(c)

        return nmea_partial + '*%02X' % checksum

    def extract_field(self, slider, size):
        field = bytearray(self.message[slider >> 3:(slider + size >> 3) + 1])
        field[0] = field[0] & (2 ** (8 - (slider & 7)) - 1)
        return int.from_bytes(field, 'big') >> (8 - (slider + size & 7))

    def set_params(self, **kwargs):
        self.__dict__.update(self.schema(**kwargs).get_params())

    def get_params(self):
        return self.__dict__



#!/usr/bin/env python3

import argparse
import logging
import socket
import sys

from ..qzss_dcr_lib.exception import QzssDcrDecoderException
from ..qzss_dcr_lib.exception import QzssDcrDecoderNotImplementedError
from ..qzss_dcr_lib.interface import decode_stream

logger = logging.getLogger(__name__)


class Transmitter:
    def __init__(self, dst_host='ff02::1', dst_port=2112, address_family=socket.AF_UNSPEC):
        self.addr_info = socket.getaddrinfo(dst_host, dst_port,
                                            address_family,
                                            socket.SOCK_DGRAM,
                                            socket.IPPROTO_UDP)[0]

    def handler(self, report):
        with socket.socket(self.addr_info[0], self.addr_info[1]) as sock:
            sat_id = (report.satellite_id or 0x55).to_bytes(1, 'big')
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            logger.info(report.nmea)
            sock.sendto(sat_id + report.message, self.addr_info[-1])

    def start(self, stream=sys.stdin, msg_type='ublox', unique=False):
        decode_stream(stream, msg_type=msg_type, callback=self.handler, unique=unique)


def main():
    logging.basicConfig(level=logging.INFO,
                        format='%(asctime)s - %(levelname)s - %(message)s')
    parser = argparse.ArgumentParser(description='azarashi network transmitter',
                                     formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-d', '--dst-host', help="destination host", type=str, default='ff02::1')
    parser.add_argument('-p', '--dst-port', help='destination port', type=int, default=2112)
    parser.add_argument('-t', '--msg-type', help="message type", type=str, choices=['hex', 'nmea', 'ublox'],
                        default='nmea')
    parser.add_argument('-f', '--input', help='input device', type=str, default='stdin')
    parser.add_argument('-u', '--unique', help='supress duplicate messages', action='store_true')
    args = parser.parse_args()
    if args.input == 'stdin':
        stream = sys.stdin
    else:
        stream = open(args.input, mode='r')

    xmitter = Transmitter(dst_host=args.dst_host, dst_port=args.dst_port)
    while True:
        try:
            xmitter.start(stream=stream, msg_type=args.msg_type, unique=args.unique)
        except QzssDcrDecoderException as e:
            logger.warning(f'[{type(e).__name__}] {e}')
        except QzssDcrDecoderNotImplementedError as e:
            logger.warning(f'[{type(e).__name__}] {e}')
        except EOFError as e:
            logger.info(f'{e}')
            break
        except Exception as e:
            logger.warning(f'[{type(e).__name__}] {e}')

    stream.close()
    return 0


if __name__ == '__main__':
    exit(main())
